import numpy as np
from micropolarray.processing.nrgf import roi_from_polar

# TODO: aggiungere tutti i bei dati qua, rendere le camere personalizzabili (caricare dark da cartelle e cose cosi)


class Camera:
    def __init__(self):
        pass

    def occulter_roi(
        self, data: np.array, fill: float = 0.0, overoccult: int = 0
    ) -> np.array:
        """Returns the array in the polar ROI, else fill

        Args:
            data (np.array): Input array
            fill (float, optional): Value for filling. Defaults to 0.0.
            overoccult (int, optional): Pixels to overoccult. Defaults to 0.

        Returns:
            np.array: Array if in ROI, fill elsewhere
        """
        y, x, r = self.occulter_pos_last
        roidata = roi_from_polar(
            data, [y, x], [r + overoccult, 5000], fill=fill
        )

        return roidata

    def occulter_mask(self, overoccult: int = 0, rmax: int = None) -> np.array:
        """Returns an array of True inside the roi, False elsewhere. Useful for mean/std operations (where=occulter_mask).

        Args:
            overoccult (int, optional): Pixels to overoccult. Defaults to 15.
            rmax (int, optional): Maximum r of the ROI. Defaults to image nearest border.

        Returns:
            np.array: Boolean roi array
        """
        y, x, r = self.occulter_pos_last
        r = r + overoccult
        if rmax is None:
            rmax = np.min([y, x])
        occulter_mask = np.where(
            roi_from_polar(
                np.ones(shape=(self.h_image, self.w_image)), [y, x], [r, rmax]
            )
            != 0,
            True,
            False,
        )
        return occulter_mask


class Kasi(Camera):
    def __init__(self):
        # self.angle_dic = {-45: 0, 0: 1, 90: 2, 45: 3}  # old
        self.angle_dic = {0: 0, 45: 1, -45: 2, 90: 3}  # new
        self.linearity_range = [0.0, 2500.0]
        self.PTC = 2.64  # [e-/ADU]
        self.readout_noise = 10  # [e-]
        self.full_well = 10500  # [e-]
        self.h_image = 3000
        self.w_image = 4096


class PolarCam(Camera):
    def __init__(self):
        # self.angle_dic = {90: 0, 45: 1, -45: 2, 0: 3}  # My ref system
        self.angle_dic = {
            0: 0,
            45: 1,
            -45: 2,
            90: 3,
        }  # Ale ref system, 0 = vertical
        self.PolarCam_model = "U4"
        self.sensor_type = "CCD"
        self.h_image = 1952  # height [pixel]
        self.w_image = 1952  # width  [pixel]
        self.pixeldim_l1 = 7.4  # [micron]
        self.pixeldim_l2 = 7.4  # [micron]
        self.orientation00 = 0  # |  [degree]
        self.orientation01 = 45  # /  [degree]
        self.orientation11 = 90  # -- [degree]
        self.orientation10 = 135  # \  [degree]
        self.saturationCapacity = 44.0e3  # [e-]
        self.bitdepth = 12  # [bit]
        self.frameRate = 14  # [fps]
        self.Texp_min = 0.02  # [ms]
        self.quantumEff = 0.76  # @470 (nominal, by user manual)
        self.occulter_pos_last = [
            919,
            950,
            531,
        ]  # Occulter y, x, radius [px, px, px] updated January 2023
        self.occulter_pos_2021 = [
            925,
            934,
            532,
        ]  # updated march 27, 2021/2022 campaign
        self.occulter_radius_sr = 1.1901  # occulter dimension in solar radii
        self.sun_dimension_pixels = 446  # from standard astropy atan(R_sun/AU)

        self.gain = 9.28


class Antarticor:
    # AntarctiCor
    def __init__(self):
        self.aperture = 50  # [mm]
        self.effectiveFocalLength = 700  # [mm]
        self.fratio = self.effectiveFocalLength / self.aperture
        self.spectralRange = 591  # (591 +- 5) [nm]
        self.platescale = 4.3  # [arcsec/superpixel]
        self.FoV_degree = 0.6  # +/- 0.6 [degree]
        self.FoV_Rsun = 2.24  # +/- 2.24 [solar radii]

        # Other
        self.Topal = 0.27  # opal transmittance []
        self.const = 1.083 * 10 ** (-5)  # constant []



import numpy as np


"""----------------------------------------------------------------------"""
""" 
    Functions to compute Angle of Linear Polarization (AoLP), Degree of
    Linear Polarization (DoLP) and Polarized Brightness (pB), returned as 
    numpy arrays.
    
    Args:
        Stokes_vec_components (numpy.array[3, number of y pixels, number of x pixels]): array containing elements of the Stokes vector of an np.array[y, x] image, in the form [S0, S1, S2].
"""


def AoLP(Stokes_vec_components):
    """Angle of linear polarization in [rad]"""
    I, Q, U = Stokes_vec_components
    angle = 0.5 * np.arctan(
        np.divide(1.0 * U, 1.0 * Q, out=np.zeros_like(U), where=Q != 0),
        dtype=float,
    )  # avoids warning when dividing by 0
    return angle

    angle = np.where(
        Q != 0,
        0.5 * np.arctan(1.0 * U / (1.0 * Q), dtype=float),
        np.deg2rad(90.0),
    )  # set it to 90 deg when denominator explodes
    return angle


def DoLP(Stokes_vec_components):
    """Degree of linear polarization in [%]"""
    I, Q, U = Stokes_vec_components
    return np.divide(
        np.sqrt((Q * Q) + (U * U), dtype=float), I, where=(I != 0)
    )  # avoids 0/0 error


def pB(Stokes_vec_components):
    """Polarized brighness in [%]"""
    I, Q, U = Stokes_vec_components
    return np.sqrt((Q * Q) + (U * U), dtype=float)


"""----------------------------------------------------------------------"""


def normalize2pi(angles_list):
    """Normalizes the input angle list in the -90,90 range

    Args:
        angles_list (_type_): list of input angles in degrees

    Returns:
        _type_: normalized angles
    """
    if type(angles_list) is not list:
        angles_list = [
            angles_list,
        ]
    for i, angle in enumerate(angles_list):
        while angle > 90:
            angle -= 180
        while angle <= -90:
            angle += 180
        angles_list[i] = angle

    return angles_list



import sys
import warnings
from logging import info

import numpy as np
from numba import njit

from micropolarray.processing.congrid import congrid

warnings.filterwarnings(
    "ignore", category=UserWarning, append=True
)  # Warnings of fits package became annoying.


@njit  # MAKES OPERATIONS REALLY FAST, checked
def demosaicmean(data):
    """Loops over right polarization pixel location, takes 1/4 of that,
    stores it in the 2x2 superpixel.
    demo_images[0] = data[y=0, x=0]
    demo_images[1] = data[y=0, x=1]
    demo_images[2] = data[y=1, x=0]
    demo_images[3] = data[y=1, x=1]
    """
    data = 1.0 * data
    temp_data = data.copy()
    demo_images = [
        np.zeros_like(temp_data),
        np.zeros_like(temp_data),
        np.zeros_like(temp_data),
        np.zeros_like(temp_data),
    ]
    counter = 0
    for j in range(2):
        for i in range(2):
            for y_super in range(0, data.shape[0], 2):
                for x_super in range(0, data.shape[1], 2):
                    mean = data[y_super + j, x_super + i] * 0.25
                    temp_data[
                        y_super : y_super + 2, x_super : x_super + 2
                    ] = mean
            demo_images[counter] = temp_data.copy()
            counter += 1

    return demo_images


@njit
def demosaicadjacent(data):
    data = 1.0 * data
    temp_data = data.copy()
    demo_images = [
        np.zeros_like(temp_data),
        np.zeros_like(temp_data),
        np.zeros_like(temp_data),
        np.zeros_like(temp_data),
    ]
    counter = 0
    for y_pix_family in range(2):
        for x_pix_family in range(2):
            for y in range(0, data.shape[0] - 2, 2):
                for x in range(0, data.shape[1] - 2, 2):
                    temp_data[y, x] = data[y + y_pix_family, x + x_pix_family]
                    temp_data[y + 1, x] = 0.5 * (
                        data[y + y_pix_family, x + x_pix_family]
                        + data[y + y_pix_family + 2, x + x_pix_family]
                    )
                    temp_data[y, x + 1] = 0.5 * (
                        data[y + y_pix_family, x + x_pix_family]
                        + data[y + y_pix_family, x + x_pix_family + 2]
                    )
                    temp_data[y + 1, x + 1] = (
                        data[y + y_pix_family, x + x_pix_family]
                        + data[y + y_pix_family + 2, x + x_pix_family]
                        + data[y + y_pix_family, x + x_pix_family + 2]
                        + data[
                            y + y_pix_family + 2,
                            x + x_pix_family + 2,
                        ]
                    ) * 0.25
            demo_images[counter] = temp_data.copy()
            counter += 1

    return demo_images


def split_polarizations(data: np.ndarray):
    single_pol_images = np.array(
        [data[j::2, i::2] for j in range(2) for i in range(2)],
        # [
        #    self.data[0::2, 0::2], # x= 0, y = 0
        #    self.data[0::2, 1::2], # x= 1, y = 0
        #    self.data[1::2, 0::2], # x= 0, y = 1
        #    self.data[1::2, 1::2], # x= 1, y = 1
        # ],
        dtype=float,
    )

    return single_pol_images


@njit
def merge_polarizations(single_pol_images: np.ndarray):
    data = np.zeros(
        shape=(
            single_pol_images[0].shape[0] * 2,
            single_pol_images[0].shape[1] * 2,
        )
    )
    data[0::2, 0::2] = single_pol_images[0]
    data[0::2, 1::2] = single_pol_images[1]
    data[1::2, 0::2] = single_pol_images[2]
    data[1::2, 1::2] = single_pol_images[3]
    return data


def demosaic(image_data, option="adjacent"):
    """
    Returns a [4,n,m] array of polarized images, starting from a
    micropolarizer image array [n, m].
    """
    data = image_data
    temp_data = data.copy()
    demo_images = np.empty([4, data.shape[0], data.shape[1]], dtype="d")

    if option == "mean":
        info("Demosaicing (mean method)... ")
        demo_images = demosaicmean(
            np.array(image_data, dtype=float)
        )  # casting needed by numba

    elif option == "adjacent":
        """
        longer: loops over superpixel, tries to get value of right polarized pixel, except if at image boundary. Then takes the mean of adjacent cells, stores it temporarily in the 2x2 pixel, then rotates it to match the right
        """
        info("Demosaicing (adjacent method)...")
        # Adding two new columns/rows to avoid segmentationfault-like error
        stacked_data = np.vstack([image_data, image_data[-2:, :]])
        stacked_data = np.hstack([stacked_data, stacked_data[:, -2:]])
        demo_images = demosaicadjacent(stacked_data)
        demo_images = np.array(demo_images)[:, :-2, :-2]

    elif option == "spline":
        single_pol_subimages = np.array(
            [image_data[j::2, i::2] for j in range(2) for i in range(2)],
            dtype=np.float,
        )
        demo_images = [
            congrid(
                single_pol_subimages[i],
                (image_data.shape[0], image_data.shape[1]),
            )
            for i in range(4)
        ]

    else:
        raise ValueError('"option" should be one of ["mean", "adjacent"]')

    return np.array(demo_images)



import numpy as np


def decompose_2d(data, num_of_chunks):
    chunk_size_y = int(data.shape[0] / num_of_chunks)
    chunk_size_x = int(data.shape[1] / num_of_chunks)
    splitted_data = np.zeros(
        shape=(num_of_chunks * num_of_chunks, chunk_size_y, chunk_size_x)
    )
    for j in range(num_of_chunks):
        for i in range(num_of_chunks):
            splitted_data[i + num_of_chunks * j] = np.array(
                data[
                    j * (chunk_size_y) : (j + 1) * chunk_size_y,
                    i * (chunk_size_x) : (i + 1) * chunk_size_x,
                ]
            )

    return splitted_data


def recompose_2d(splitted_data, num_of_chunks):
    height, width = splitted_data[0].shape
    recomposed_data = np.zeros(
        shape=(height * num_of_chunks, width * num_of_chunks)
    )
    for j in range(num_of_chunks):
        for i in range(num_of_chunks):
            recomposed_data[
                j * (height) : (j + 1) * height,
                i * (width) : (i + 1) * width,
            ] = splitted_data[i + num_of_chunks * j]

    return recomposed_data



from __future__ import annotations

import datetime
import os
import sys
from logging import info, warning
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
from astropy.io import fits

from micropolarray.utils import fix_data, make_abs_and_create_dir


class Image:
    """Basic image class. Can be initialized from a filename, a filenames list, a numpy array or another Image instance. If multiple filenames are provided, will perform the mean of them unless averageimages is False."""

    def __init__(
        self,
        initializer: str | np.ndarray | Image,
        averageimages: bool = True,
    ):
        self.header = None
        if type(initializer) is str or type(initializer) is list:
            self._init_image_from_file(initializer, averageimages)
        elif type(initializer) is np.ndarray:
            self._init_image_from_data(initializer)
        elif type(initializer) is Image:
            self._init_image_from_image(initializer)

    def _init_image_from_data(self, data: np.array) -> None:
        self._set_data(np.array(data))
        self.filename = "image.fits"

    def _init_image_from_file(self, filenames, averageimages) -> None:
        filenames_list = (
            [filenames] if type(filenames) is not list else filenames
        )

        filenames_len = len(filenames_list)
        if filenames_len == 0:
            raise NameError("Can't load files, empty filenames list.")
        datetimes = [0] * filenames_len

        print_info_message = True
        for idx, filename in enumerate(filenames_list):
            if ".raw" in filenames_list:
                raise NameError(
                    "Can't load a raw file, convert it to fits first."
                )  # to fix
            with fits.open(filename) as hul:
                if idx == 0:
                    combined_data = hul[0].data / (
                        1 + int(averageimages) * (filenames_len - 1)
                    )
                    self.header = hul[0].header
                else:
                    if print_info_message:
                        if averageimages:
                            info(f"Averaging {filenames_len} images...")
                        else:
                            info(f"Summing {filenames_len} images...")
                        print_info_message = False
                    combined_data = combined_data + (
                        hul[0].data
                        / (1 + int(averageimages) * (filenames_len - 1))
                    )  # divide by 1 if summing, either n if averaging

                hul.verify("fix")

                try:  # standard format
                    datetimes[idx] = datetime.datetime.strptime(
                        hul[0].header["DATE-OBS"],
                        "%Y-%m-%dT%H:%M:%S",
                    )
                except ValueError:  # antarticor format
                    datetimes[idx] = datetime.datetime.strptime(
                        hul[0].header["DATE-OBS"]
                        + "-"
                        + hul[0].header["TIME-OBS"],
                        "%Y-%m-%d-%H:%M:%S",
                    )
                except KeyError:
                    pass

        datetimes = [datetime for datetime in datetimes if datetime != 0]
        if len(datetimes) == 0:
            datetimes = [0]
        self._set_data(np.array(combined_data))

        if filenames_len > 1:
            self.header["SUMOF"] = (
                filenames_len,
                "Number of files the images is the sum of.",
            )
            datetimes = sorted(datetimes)
            if not datetimes[0] == 0:
                self.header["STRT-OBS"] = (
                    str(datetimes[0]),
                    "Date and time of first image.",
                )
                self.header["END-OBS"] = (
                    str(datetimes[-1]),
                    "Date and time of last image.",
                )
            tempfilename = filenames[0].split(os.path.sep)
            tempfilename[-1] = "SUM_" + tempfilename[-1]
            self.filename = os.path.sep.join(tempfilename)
        else:
            self.filename = filenames[0]

    def _init_image_from_image(self, image: Image):
        self._set_data(image.data.copy())
        self.filename = image.filename

    def _set_data(self, data: np.array):
        """Set image data and derived polarization informations, and
        consequently change header."""
        self.data = data
        self.height, self.width = data.shape
        if self.header is None:
            self.header = self._set_default_header(data)
        else:
            self._update_dims_in_header(self.data)

    def _set_default_header(self, data: np.array):
        """Set a default header when initializing image with data instead of
        .fits filename."""
        header = fits.PrimaryHDU(data=data).header
        return header

    def _update_dims_in_header(self, data: np.array):
        self.header["NAXIS1"] = data.shape[1]
        self.header["NAXIS2"] = data.shape[0]

    # ----------------------------------------------------------------
    # -------------------------- SAVING ------------------------------
    # ----------------------------------------------------------------

    def save_as_fits(self, filename: str, fixto: str[float, float] = None):
        """Saves image as fits with current header

        Args:
            filename (str): output filename
            fixto (str[float, float], optional): Set a maximum and minimum range for the data. Defaults to None.

        Raises:
            ValueError: filename does not end with ".fits"
        """
        filepath = Path(filename)
        if filepath.suffix != ".fits":
            print(filepath.suffix)
            raise ValueError("filename must be a valid file name, not folder.")
        filepath = Path(make_abs_and_create_dir(filename))
        if fixto:
            data = fix_data(self.data, *fixto)
        else:
            data = self.data
        hdu = fits.PrimaryHDU(
            data=data,
            header=self.header,
            do_not_scale_image_data=True,
            uint=False,
        )
        filename = str(
            filepath.joinpath(filepath.parent, filepath.stem + ".fits")
        )
        hdu.writeto(filename, overwrite=True)
        info(f'Image successfully saved to "{filename}".')

    def save_as_raw(self, filename: str):
        """Saves the image as a raw binary file

        Args:
            filename (str): output filename

        Raises:
            ValueError: filename does not end with ".raw"
        """
        print(self.data.shape)
        if Path(filename).suffix != ".raw":
            raise ValueError("Filename must have .raw extension")
        self.data.astype("int16").tofile(filename)

    # ----------------------------------------------------------------
    # ------------------------------ SHOW ----------------------------
    # ----------------------------------------------------------------

    def show(self, cmap="Greys_r", vmin=None, vmax=None) -> tuple:
        """Shows the image data

        Args:
            cmap (str, optional): figure colorbar. Defaults to "Greys_r".
            vmin (_type_, optional): Minimum value to plot. Defaults to None.
            vmax (_type_, optional): Maximum value to plot. Defaults to None.

        Returns:
            tuple: fig, ax tuple as returned by matplotlib.pyplot.subplots
        """
        data_to_plot = self.data
        data_ratio = data_to_plot.shape[0] / data_to_plot.shape[1]
        # image_fig, imageax = plt.subplots(figsize=(9, 9))
        image_fig, imageax = plt.subplots(dpi=200)
        if vmin is None:
            vmin = np.min(data_to_plot)
        if vmax is None:
            vmax = np.max(data_to_plot)
        pos = imageax.imshow(data_to_plot, cmap=cmap, vmin=vmin, vmax=vmax)
        imageax.set_title(
            f"Image data (avrg {np.mean(data_to_plot, where=np.where(data_to_plot>0, True, False)):3.2f}+-{np.std(data_to_plot, where=np.where(data_to_plot>0, True, False)):3.2f})",
            color="black",
        )
        imageax.set_xlabel("x")
        imageax.set_ylabel("y")
        image_fig.colorbar(
            pos, ax=imageax, label="DN", fraction=data_ratio * 0.05
        )

        return image_fig, imageax

    def show_histogram(self, bins: int = 1000) -> tuple:
        """Print the histogram of the flattened image data

        Args:
            bins (int, optional): Numbers of bin to compute the histogram from. Defaults to 1000.

        Returns:
            tuple: fig, ax tuple as returned by matplotlib.pyplot.subplots
        """
        fig, ax = plt.subplots(dpi=200, constrained_layout=True)
        histo = np.histogram(self.data, bins=bins)
        ax.stairs(*histo)
        ax.set_title("Image histogram", color="black")
        ax.set_xlabel("Signal [DN]")
        ax.set_ylabel("Number of pixels")

        return fig, ax

    def __add__(self, second) -> Image:
        if type(self) is type(second):
            newdata = self.data + second.data
        else:
            newdata = self.data + second
        return Image(newdata)

    def __sub__(self, second) -> Image:
        if type(self) is type(second):
            newdata = self.data - second.data
        else:
            newdata = self.data - second
        return Image(newdata)

    def __mul__(self, second) -> Image:
        if type(self) is type(second):
            newdata = self.data * second.data
        else:
            newdata = self.data * second
        return Image(newdata)

    def __truediv__(self, second) -> Image:
        if type(self) is type(second):
            newdata = np.divide(self.data, second.data, where=second.data != 0)
        else:
            newdata = np.divide(self.data, second, where=second != 0)
        return Image(newdata)



import numpy as np
from numba import njit
from micropolarray.processing.demosaic import (
    split_polarizations,
    merge_polarizations,
)


@njit
def shift(data: np.ndarray, y: int, x: int):
    newdata = np.zeros_like(data)
    for j in range((y > 0) * y, newdata.shape[0] - np.abs(y) * (y < 0)):
        for i in range((x > 0) * x, newdata.shape[1] - np.abs(x) * (x < 0)):
            newdata[j, i] = data[j - y, i - x]

    return newdata


def shift_micropol(data: np.ndarray, y: int, x: int):
    """Splits the image into single polarizations, shifts each of them by y,x and then merges them back.

    Args:
        data (np.ndarray): array to shift
        y (int): vertical shift (positive inside the image)
        x (int): horizontal shift (positive inside the image)

    Returns:
        np.ndarray: shifted array
    """
    single_pol_subimages = split_polarizations(data)
    new_single_pol_subimages = np.zeros_like(single_pol_subimages)

    for i, image in enumerate(single_pol_subimages):
        new_single_pol_subimages[i, :, :] = shift(image, y, x)

    return merge_polarizations(new_single_pol_subimages)



import numpy as np
import scipy

from micropolarray.micropol_image import MicropolImage
from micropolarray.processing.demosaic import merge_polarizations


def get_hot_pixels(image, threshold=100):
    subimages = image.single_pol_subimages
    blurred_subimages = np.array(
        [
            scipy.ndimage.median_filter(subimage, size=2)
            for subimage in subimages
        ]
    )
    contrast = (subimages - blurred_subimages) / (
        subimages + blurred_subimages
    )
    diff = np.where(contrast > threshold, 1, 0)
    # diff = subimages - blurred_subimages
    # diff = np.where(diff > threshold, 1, 0)

    newimage = MicropolImage(image)
    newimage._set_data_and_Stokes(merge_polarizations(diff))

    return newimage



import glob
import os
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import pytest
from astropy.io import fits
from scipy.optimize import curve_fit

import micropolarray as ml
from micropolarray.processing.demodulation import Malus


@pytest.fixture(autouse=True)
def dummy_data():
    """Dummy data factory"""

    def _make_dummy_data(dimension):
        dummydata = np.zeros(shape=(dimension, dimension))
        dummydata[0::2, 0::2] = 1
        dummydata[0::2, 1::2] = 2
        dummydata[1::2, 0::2] = 3
        dummydata[1::2, 1::2] = 4
        return dummydata

    return _make_dummy_data


def write_temp_image(tmp_path, data):
    """Writes images to the temp folder for testing"""
    image = fits.PrimaryHDU(
        data=data, do_not_scale_image_data=True, uint=False
    )
    image.writeto(tmp_path / "sample_image.fits")


def test_image_initialization(dummy_data, tmp_path):
    """Tests the initialization of both Image and MicroPolArrayImage"""
    dummy_data_16 = dummy_data(16)
    write_temp_image(tmp_path, dummy_data_16)
    for image_type in [ml.Image, ml.MicropolImage]:
        image = image_type(dummy_data_16)
        assert np.all(image.data == dummy_data_16)

        image = image_type(str(tmp_path / "sample_image.fits"))
        assert np.all(image.data == dummy_data_16)

        image = image_type(image)
        assert np.all(image.data == dummy_data_16)


def test_image_writing(dummy_data, tmp_path):
    """Tests the saving of both Image and MicroPolArrayImage"""
    dummy_data_16 = dummy_data(16)
    for image_type in [ml.Image, ml.MicropolImage]:
        image = image_type(dummy_data_16)
        image.save_as_fits(str(tmp_path / "image.fits"))


def test_demosaic(dummy_data, tmp_path):
    """Tests demosaic operation and demosaic writing"""
    dummy_data_16 = dummy_data(16)
    # test mean
    image = ml.MicropolImage(dummy_data_16, demosaic_mode="mean")
    image = image.demosaic()
    for idx, demo_image in enumerate(image.demosaiced_images):
        assert np.all(demo_image == np.full((16, 16), (idx + 1) / 4.0))

    # test adjacent
    image = ml.MicropolImage(dummy_data_16, demosaic_mode="adjacent")
    image = image.demosaic()
    for idx, demo_image in enumerate(image.demosaiced_images):
        assert np.all(demo_image == np.full((16, 16), (idx + 1)))

    # test writing
    image.save_demosaiced_images_as_fits(
        str(tmp_path / "demosaiced_images.fits")
    )


def test_rebinning(dummy_data):
    """Tests 2x2 and 4x4 binning (the other will be supposedly fine)"""
    dummy_data_16 = dummy_data(16)

    binned_image_2 = ml.MicropolImage(dummy_data_16).rebin(2)
    assert np.all(binned_image_2.data == dummy_data(8) * 4)

    binned_image_4 = ml.MicropolImage(dummy_data_16).rebin(4)
    assert np.all(binned_image_4.data == dummy_data(4) * 16)


def test_pol_parameters(dummy_data):
    """Test if polarization parameters are correcly computed"""

    def test_theo_stokes(image, I, Q, U):
        assert np.all(image.I.data == I)
        assert np.all(image.Q.data == Q)
        assert np.all(image.U.data == U)

        assert np.all(image.AoLP.data == 0.5 * np.arctan(U / Q) * half_ones)
        assert np.all(image.pB.data == np.sqrt(Q * Q + U * U) * half_ones)
        assert np.all(
            image.DoLP.data == np.sqrt(Q * Q + U * U) * half_ones / I
        )

    array_side = 16
    dummy_data_16 = dummy_data(array_side)
    half_ones = np.ones(shape=(int(array_side / 2), int(array_side / 2)))
    image = ml.MicropolImage(dummy_data_16)

    for i in range(4):
        assert np.all(image.single_pol_subimages[i] == i + 1)

    angles = [0, 45, -45, 90]
    numbers = [1.0, 2.0, 3.0, 4.0]
    for angle, n in zip(angles, numbers):
        assert np.all(image.single_pol_subimages[image.angle_dic[angle]] == n)

    I = 0.5 * (1 + 2 + 3 + 4)
    Q = 1.0 - 4.0
    U = 2.0 - 3.0
    test_theo_stokes(image, I, Q, U)

    new_angle_dic = {45: 0, 0: 1, 90: 2, -45: 3}
    image = ml.MicropolImage(dummy_data_16, angle_dic=new_angle_dic)
    assert image.angle_dic == new_angle_dic

    Q = 2.0 - 3.0
    U = 1.0 - 4.0
    test_theo_stokes(image, I, Q, U)


# TODO
def test_demodulation(dummy_data, tmp_path):
    """Create a dummy demodulation matrix, save it, read it then use it to demodulate. Check if demodulation is correctly done."""
    dummy_data_16 = dummy_data(16)
    angles = np.array([np.deg2rad(angle) for angle in [0, 45, -45, 90]])
    demo_matrix = np.array(
        [
            [0.5, 0.5, 0.5, 0.5],
            np.cos(2 * angles),
            np.sin(2 * angles),
        ]
    )
    for i in range(3):
        for j in range(4):
            image = fits.PrimaryHDU()
            image.data = np.ones_like(dummy_data_16) * demo_matrix[i, j]
            image.writeto(tmp_path / f"M{i}{j}.fits")

    image = ml.PolarcamImage(dummy_data_16)
    assert np.all(image.Q.data == (1 - 4))
    assert np.all(image.U.data == (2 - 3))
    assert np.all(image.I.data == (0.5 * (1 + 2 + 3 + 4)))

    demodulator = ml.Demodulator(str(tmp_path))

    demo_image = image.demodulate(demodulator=demodulator)
    assert np.all(np.round(demo_image.Q.data, 5) == (1.0 - 4.0))
    assert np.all(np.round(demo_image.U.data, 5) == (2.0 - 3.0))
    assert np.all(
        np.round(demo_image.I.data, 5) == (0.5 * (1.0 + 2.0 + 3.0 + 4.0))
    )

    # try demodulation
    output_dir = tmp_path / "computed_matrix"
    output_str = str(output_dir)

    polarizations = np.arange(-45, 91, 15)
    pols_rad = np.deg2rad(polarizations)
    input_signal = 100
    eff = 0.7
    t = 0.9
    side = 10
    ones = np.ones(shape=(side, side))
    for pol, pol_rad in zip(polarizations, pols_rad):
        subimages = np.array(
            [
                ones * input_signal * Malus(pol_rad, t, eff, angle)
                for angle in angles
            ]
        )
        result_image = ml.MicropolImage(ml.merge_polarizations(subimages))
        result_image.save_as_fits(tmp_path / f"pol_{int(pol)}.fits")
    if False:  # check that fit will be ok
        for angle in angles:
            pars, pcov = curve_fit(
                Malus,
                pols_rad,
                np.array([Malus(pol, t, eff, angle) for pol in pols_rad]),
            )
            print(f"t = {pars[0]}")
            print(f"eff = {pars[1] }")
            print(f"phi = {np.rad2deg(pars[2]) }")

    # read the files
    filenames = sorted(
        glob.glob(str(tmp_path / "pol*.fits")),
        key=lambda x: int(x.split(os.path.sep)[-1][4:].strip(".fits")),
    )

    ml.calculate_demodulation_tensor(
        polarizer_orientations=polarizations,
        filenames_list=filenames,
        micropol_phases_previsions=angles,
        gain=2.75,
        output_dir=output_str,
        binning=1,
        procs_grid=[2, 2],
        normalizing_S=input_signal,
        DEBUG=False,
    )

    example_pols = np.array([ones * input_signal * 0.5 for i in range(4)])
    example_image = ml.MicropolImage(ml.merge_polarizations(example_pols))

    assert np.all(example_image.I.data == input_signal)
    assert np.all(example_image.Q.data == 0)
    assert np.all(example_image.U.data == 0)
    assert np.all(example_image.DoLP.data == 0)
    assert np.all(example_image.AoLP.data == 0)
    assert np.all(example_image.pB.data == 0)

    demodulator = ml.Demodulator(output_str)
    example_image = example_image.demodulate(demodulator)
    if False:
        demodulator.show()
        example_image.show_with_pol_params()
        plt.show()

    # assert np.all(np.round(example_image.I.data, 5) == 100) # AAAAAAAAA probably not working because of penrose inversion instead of inv
    assert np.all(np.round(example_image.Q.data, 5) == 0)
    assert np.all(np.round(example_image.U.data, 5) == 0)
    assert np.all(np.round(example_image.DoLP.data, 5) == 0)
    assert np.all(np.round(example_image.AoLP.data, 5) == 0)
    assert np.all(np.round(example_image.pB.data, 5) == 0)

    """
        data = ml.merge_polarizations()
        image = ml.MicropolImage()
    """



from numba import njit
import numpy as np
from scipy.interpolate import griddata
import scipy.interpolate
import scipy.ndimage


# not working
@njit
def micropolarray_jitcongrid(data, width, height, scale):
    new_width = int(width * scale)
    new_height = int(height * scale)
    new_data = np.zeros(shape=(new_height, new_width), dtype=float)
    meshgrid = np.zeros(shape=(new_height, new_width, 2))
    centers = [
        [np.sqrt((0.5 + i) ** 2 + (0.5 + j) ** 2) for i in range(new_height)]
        for j in range(new_width)
    ]

    # Fill grid with nearest points
    for new_i in range(new_width):
        for new_j in range(new_height):
            pass
    return centers


def interpolate(x, x_0, y_0, x_1, y_1):
    return (x - x_0) * ((y_1 - y_0) / (x_1 - x_0)) + y_0


# Copied from IDL congrid, arbitrary reshape
def congrid(a, newdims, kind="linear") -> np.ndarray:
    """Reshapes the data into any new lenght and width

    Args:
        a (np.array): data to be reshaped
        newdims (tuple | list): new lenght and width
        kind (str, optional): interpolation type. Defaults to "linear".

    Returns:
        ndarray: numpy array of congridded image
    """
    if not a.dtype in [np.float64, np.float32]:
        a = np.cast[float](a)

    m1 = np.cast[int](True)
    old = np.array(a.shape)
    ndims = len(a.shape)
    newdims = np.asarray(newdims, dtype=float)
    dimlist = []

    # Linear method, default
    for i in range(ndims):
        base = np.arange(newdims[i])
        dimlist.append((old[i] - m1) / (newdims[i] - m1) * base)

    olddims = [np.arange(i, dtype=np.float) for i in list(a.shape)]
    mint = scipy.interpolate.interp1d(
        olddims[-1], a, kind=kind, fill_value="extrapolate"
    )
    newa = mint(dimlist[-1])

    trorder = [ndims - 1] + list(range(ndims - 1))
    for i in range(ndims - 2, -1, -1):
        newa = newa.transpose(trorder)
        mint = scipy.interpolate.interp1d(
            olddims[i], newa, kind=kind, fill_value="extrapolate"
        )
        newa = mint(dimlist[i])
    if ndims > 1:
        newa = newa.transpose(trorder)

    return newa



from __future__ import annotations

from dataclasses import dataclass
from logging import debug, error, info, warning
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import scipy
from astropy.io import fits
from PIL import Image as PILImage

from micropolarray.cameras import Camera, PolarCam
from micropolarray.image import Image
from micropolarray.polarization_functions import AoLP, DoLP, pB
from micropolarray.processing.chen_wan_liang_calibration import ifov_jitcorrect
from micropolarray.processing.congrid import congrid
from micropolarray.processing.demodulation import Demodulator
from micropolarray.processing.demosaic import (
    demosaic,
    merge_polarizations,
    split_polarizations,
)
from micropolarray.processing.nrgf import roi_from_polar
from micropolarray.processing.rebin import micropolarray_rebin
from micropolarray.processing.shift import shift, shift_micropol
from micropolarray.utils import (
    fix_data,
    make_abs_and_create_dir,
    mean_minus_std,
    mean_plus_std,
)


@dataclass
class PolParam:
    """Auxiliary class for polarization parameters.

    Members:
        ID (str): parameter identifier
        data (np.array): parameter image as numpy 2D array
        title (str): brief title of the parameter, useful for plotting
        measure_unit (str): initial measure units of the parameter
        fix_data (bool): controls whether data has to be constrained to [0, 4096] interval (not implemented yet)
    """

    ID: str
    data: np.ndarray
    title: str
    measure_unit: str
    fix_data: bool


DEFAULT_ANGLES_DIC = None  # sets the micropolarizer orientations with a dictionary {angle : position in superpix 1->3}


def set_default_angles(angles_dic: dict):
    """Sets the default micropolarizer orientations for images.

    Args:
        angles_dic (dict): dictionary {value : pos} where value is the angle in degrees from -90 to 90 and pos is the pixel position in superpixel, from 0 to 3 (position [y, x], fast index x)
    """
    global DEFAULT_ANGLES_DIC
    DEFAULT_ANGLES_DIC = angles_dic


class MicropolImage(Image):
    """Micro-polarizer array image class. Can be initialized from a 2d array, a list of 1 or more file names (use the boolean keyword averageimages to select if sum or average is taken) or another MicropolImage. Dark and flat micropolarray images can also be provided to automatically correct the result."""

    first_call = True  # Avoid repeating messages

    def __init__(
        self,
        initializer: str | np.ndarray | list | MicropolImage,
        angle_dic: dict = None,
        demosaic_mode: str = "adjacent",
        dark: MicropolImage = None,
        flat: MicropolImage = None,
        averageimages: bool = True,
    ):
        self._is_demodulated = False
        self._is_demosaiced = False
        self._binning = 1
        self._flat_subtracted = False
        self._dark_subtracted = False
        if angle_dic is None:
            global DEFAULT_ANGLES_DIC
            if DEFAULT_ANGLES_DIC is None:
                if MicropolImage.first_call:
                    warning(
                        f"Micropolarizer orientation dictionary defaults to {PolarCam().angle_dic}, set it via set_default_angles(camera)\n"
                    )
                MicropolImage.first_call = False
                DEFAULT_ANGLES_DIC = PolarCam().angle_dic
            angle_dic = DEFAULT_ANGLES_DIC
        self.angle_dic = angle_dic
        self.demosaic_mode = demosaic_mode
        self.demosaiced_images = None

        if type(initializer) is str and len(initializer) > 1:
            self._num_of_images = len(initializer)
        else:
            self._num_of_images = 1
        super().__init__(
            initializer=initializer, averageimages=averageimages
        )  # Call generic Image() constructor

        if (type(initializer) is list) or (type(initializer) is str):
            self._init_micropolimage_from_file(initializer)
        elif type(initializer) is np.ndarray:
            self._init_micropolimage_from_data(initializer)
        elif type(initializer) is MicropolImage:
            self._init_micropolimage_from_image(initializer)

        self._update_stokes_derived_internal_dataclasses()

        # Apply corrections if needed
        if dark is not None:
            self.subtract_dark(dark=dark)
        if flat is not None:
            self.correct_flat(flat=flat)
        elif MicropolImage.first_call:
            warning("Remember to set dark")
            MicropolImage.first_call = False

        self.height, self.width = self.data.shape

    def _init_micropolimage_from_file(self, filenames: str):
        self._set_data_and_Stokes(self.data)

    def _init_micropolimage_from_data(self, data: np.array):
        self._set_data_and_Stokes(self.data)
        if self.header is None:
            self.header = self.set_default_header(data)
        else:
            self._update_dims_in_header(self.data)

    def _init_micropolimage_from_image(self, image: MicropolImage):
        self._is_demodulated = image._is_demodulated
        self._is_demosaiced = image._is_demosaiced
        self._binning = image._binning
        self._dark_subtracted = image._dark_subtracted
        self._flat_subtracted = image._flat_subtracted
        self.angle_dic = image.angle_dic
        self.demosaic_mode = image.demosaic_mode

        self.data = image.data
        self.header = image.header
        self.single_pol_subimages = image.single_pol_subimages
        self.demosaiced_images = image.demosaiced_images
        self.Stokes_vec = image.Stokes_vec
        self.polparam_list = image.polparam_list

    # ----------------------------------------------------------------
    # ---------------------- STOKES COMPONENTS -----------------------
    # ----------------------------------------------------------------

    def _update_Stokes_vec(self) -> None:
        if not self._is_demosaiced:
            self.Stokes_vec = self._get_theo_Stokes_vec_components(
                self.single_pol_subimages
            )
        else:
            self.Stokes_vec = self._get_theo_Stokes_vec_components(
                self.demosaiced_images
            )

    def _update_stokes_derived_internal_dataclasses(self) -> None:
        self.I = PolParam(
            "I", self.Stokes_vec[0], "Stokes I", "DN", fix_data=False
        )
        self.Q = PolParam(
            "Q", self.Stokes_vec[1], "Stokes Q", "DN", fix_data=False
        )
        self.U = PolParam(
            "U", self.Stokes_vec[2], "Stokes U", "DN", fix_data=False
        )
        self.pB = PolParam(
            "pB",
            pB(self.Stokes_vec),
            "Polarized brightness",
            "DN",
            fix_data=False,
        )
        self.AoLP = PolParam(
            "AoLP",
            AoLP(self.Stokes_vec),
            "Angle of linear polarization",
            "rad",
            fix_data=False,
        )
        self.DoLP = PolParam(
            "DoLP",
            DoLP(self.Stokes_vec),
            "Degree of linear polarization",
            "",
            fix_data=False,
        )
        self.polparam_list = [
            self.I,
            self.Q,
            self.U,
            self.pB,
            self.AoLP,
            self.DoLP,
        ]

    def demodulate(self, demodulator: Demodulator) -> MicropolImage:
        """Returns a MicropolImage with polarization parameters calculated from the demodulation tensor provided.

        Args:
            demodulator (Demodulator): Demodulator object containing the demodulation tensor components (see processing.new_demodulation)

        Raises:
            ValueError: raised if image and demodulator do not have the same dimension, for example in case of different binning

        Returns:
            MicropolImage: copy of the input image with I, Q, U, pB, DoLP, AoLP calculated from the demodulation tensor.
        """
        if (self.height, self.width) != (
            demodulator.mij.shape[2],
            demodulator.mij.shape[3],
        ):
            print(demodulator.mij.shape[2], demodulator.mij.shape[3])
            print(self.height, self.width)
            raise ValueError(
                "Image and demodulator do not have the same dimensions, check binning."
            )
        info("Demodulating...")
        demodulated_image = MicropolImage(self)
        demodulated_image.Stokes_vec = (
            demodulated_image._get_Stokes_from_demodulator(demodulator)
        )
        demodulated_image._update_single_pol_subimages()
        demodulated_image._update_stokes_derived_internal_dataclasses()
        demodulated_image._is_demodulated = True

        info("Image correctly demodulated")
        return demodulated_image

    def set_data_only(self, data: np.array = None) -> None:
        if data is None:
            data = self.data
        self.data = data
        self.height = data.shape[0]
        self.width = data.shape[1]
        if (data.shape[0] % 2) or (data.shape[1] % 2):
            warning(
                "Odd number of pixels is incompatible"
                " with micropolarizer arrays operations."
            )
        self._update_single_pol_subimages()
        if self._is_demosaiced:
            self.demosaic()
        self._update_Stokes_vec()
        self._update_stokes_derived_internal_dataclasses()
        if self.header is None:
            self.header = self.set_default_header(data)
        else:
            self._update_dims_in_header(self.data)

    def _set_data_and_Stokes(self, data: np.array = None) -> None:
        """Set image data and derived polarization informations, and
        consequently change header."""
        if data is None:
            data = self.data
        self.data = data
        if (data.shape[0] % 2) or (data.shape[1] % 2):
            warning(
                "Odd number of pixels is incompatible"
                " with micropolarizer arrays operations."
            )
        self.height, self.width = data.shape
        self._update_single_pol_subimages()

        if not self._is_demosaiced:
            self.Stokes_vec = self._get_theo_Stokes_vec_components(
                self.single_pol_subimages
            )
            self._update_stokes_derived_internal_dataclasses()
        else:
            self.demosaic()
            self.Stokes_vec = self._get_theo_Stokes_vec_components(
                self.demosaiced_images
            )

    def _update_single_pol_subimages(self) -> None:
        single_pol_subimages = split_polarizations(self.data)

        self.single_pol_subimages = single_pol_subimages
        self.pol0 = PolParam(
            "0",
            single_pol_subimages[self.angle_dic[0]],
            "0 deg orientation pixels",
            "DN",
            fix_data=False,
        )
        self.pol45 = PolParam(
            "45",
            single_pol_subimages[self.angle_dic[45]],
            "45 deg orientation pixels",
            "DN",
            fix_data=False,
        )
        self.pol_45 = PolParam(
            "-45",
            single_pol_subimages[self.angle_dic[-45]],
            "-45 deg orientation pixels",
            "DN",
            fix_data=False,
        )
        self.pol90 = PolParam(
            "90",
            single_pol_subimages[self.angle_dic[90]],
            "90 deg orientation pixels",
            "DN",
            fix_data=False,
        )

    def _get_theo_Stokes_vec_components(self, single_pol_images) -> np.array:
        """
        Computes stokes vector components from four polarized images at four angles, angle_dic describes the coupling between
        poled_images_array[i] <--> angle_dic[i]
        Return:
            stokes vector, shape=(3, poled_images.shape[1], poled_images.shape[0])
        """
        I = 0.5 * np.sum(single_pol_images, axis=0)
        Q = (
            single_pol_images[self.angle_dic[0]]
            - single_pol_images[self.angle_dic[90]]
        )
        U = (
            single_pol_images[self.angle_dic[45]]
            - single_pol_images[self.angle_dic[-45]]
        )

        S = np.array([I, Q, U], dtype=float)
        return S

    def _get_Stokes_from_demodulator(
        self, demodulator: Demodulator
    ) -> np.array:
        """
        Computes stokes vector components from four polarized images at four angles, angle_dic describes the coupling between
        poled_images_array[i] <--> angle_dic[i]. Assumes:

        I = M_00 * I_1 + M_01 * I_2 + M_02 * I_3 + M_03 * I_4
        Q = M_10 * I_1 + M_11 * I_2 + M_12 * I_3 + M_13 * I_4
        U = M_20 * I_1 + M_21 * I_2 + M_22 * I_3 + M_23 * I_4

        Return:
            stokes vector, shape=(3, poled_images.shape[1], poled_images.shape[0])
        """
        # Corrected with demodulation matrices, S.shape = (4, n, m)
        num_of_malus_parameters = 3  # 3 multiplication params
        pixels_in_superpix = 4
        mij = demodulator.mij

        self.demosaic()
        demosaiced_images = self.demosaiced_images
        # demosaiced_images = demosaic(self.data, self.demosaic_mode)

        # IMG = np.array(
        #    [
        #        demosaiced_images[self.angle_dic[0]],
        #        demosaiced_images[self.angle_dic[45]],
        #        demosaiced_images[self.angle_dic[-45]],
        #        demosaiced_images[self.angle_dic[90]],
        #    ],
        #    dtype=float,
        # )  # Liberatore article/thesis

        IMG = np.array(
            [demo_image for demo_image in demosaiced_images],
            dtype=float,
        )
        if (mij[0, 0].shape[0] != IMG[0].shape[0]) or (
            mij[0, 0].shape[1] != IMG[0].shape[1]
        ):
            raise ValueError(
                f"demodulation matrix and demosaiced images have different shape {mij[0,0].shape} {IMG[0].shape}. Check that binning is correct."
            )  # sanity check

        T_ij = np.zeros(
            shape=(
                num_of_malus_parameters,
                pixels_in_superpix,
                *self.data.shape,
            )
        )
        for i in range(num_of_malus_parameters):
            for j in range(pixels_in_superpix):
                temp_tij = np.multiply(
                    demodulator.mij[i, j, :, :], IMG[j, :, :]
                )  # Matrix product
                T_ij[i, j, :, :] = temp_tij

        I = T_ij[0, 0] + T_ij[0, 1] + T_ij[0, 2] + T_ij[0, 3]
        Q = T_ij[1, 0] + T_ij[1, 1] + T_ij[1, 2] + T_ij[1, 3]
        U = T_ij[2, 0] + T_ij[2, 1] + T_ij[2, 2] + T_ij[2, 3]

        S = np.array([I, Q, U], dtype=float)
        return S

    def subtract_dark(self, dark: MicropolImage) -> MicropolImage:
        """Correctly subtracts the input dark image from the image

        Args:
            dark (MicropolImage): dark to subtract

        Returns:
            MicropolImage: copy of input image with dark subtracted
        """
        self.data = self.data - dark.data
        self.data = np.where(self.data >= 0, self.data, 0)  # Fix data
        self._set_data_and_Stokes()
        self._dark_subtracted = True
        return self

    def correct_flat(self, flat: MicropolImage) -> MicropolImage:
        """Normalizes the flat and uses it to correct the image.

        Args:
            flat (MicropolImage): flat image, does not need to be normalized.

        Returns:
            MicropolImage: copy of input image corrected by flat
        """
        normalized_flat = flat.data / np.max(flat.data)

        self.data = np.divide(
            self.data,
            normalized_flat,
            where=normalized_flat != 0.0,
        )

        # self.data = np.where(self.data >= 0, self.data, 0)
        # self.data = np.where(self.data < 4096, self.data, 4096)
        self._set_data_and_Stokes()
        self._flat_subtracted = True
        return self

    def correct_ifov(self) -> MicropolImage:
        """Corrects differences in single pixels fields of view inside each superpixel

        Returns:
            MicropolImage: image with data corrected for field of view differences
        """
        corrected_data = self.data.copy()
        corrected_data = ifov_jitcorrect(self.data, self.height, self.width)
        self._set_data_and_Stokes(corrected_data)
        return self

    # ----------------------------------------------------------------
    # ------------------------------ SHOW ----------------------------
    # ----------------------------------------------------------------

    def show_with_pol_params(self, cmap="Greys_r") -> tuple:
        """Returns a fig for each set of image parameters. User must call
        plt.show after this is called.
        Returned parameters:
        - Original image
        - Stokes vector I, Q, U
        - Angle, degree of linear polarizaimagetion Polarized brightness


        Args:
            cmap (str, optional): colormap string. Defaults to "Greys_r".

        Returns:
            tuple: a (figure, axis) couple same as matplotlib.pyplot.subplots for the image data and another for the six polarization parameters
        """
        data_ratio = self.data.shape[0] / self.data.shape[1]
        image_fig, imageax = plt.subplots(dpi=200, constrained_layout=True)
        mappable = imageax.imshow(
            self.data,
            cmap=cmap,
            vmin=mean_minus_std(self.data),
            vmax=mean_plus_std(self.data),
        )
        avg = np.mean(self.data)
        stdev = np.std(self.data)
        imageax.set_title(
            f"Image data (avrg {avg:3.2f}+-{stdev:3.2f})",
            color="black",
        )
        imageax.set_xlabel("x [px]")
        imageax.set_ylabel("y [px]")
        image_fig.colorbar(
            mappable, ax=imageax, label="[DN]", fraction=data_ratio * 0.05
        )
        stokes_fig, stokesax = plt.subplots(
            2, 3, figsize=(14, 9), constrained_layout=True
        )

        stokesax = stokesax.ravel()
        for i, stokes in enumerate(stokesax):
            vmin = mean_minus_std(self.polparam_list[i].data)
            vmax = mean_plus_std(self.polparam_list[i].data)
            avg = np.mean(self.polparam_list[i].data)
            stdev = np.std(self.polparam_list[i].data)
            mappable_stokes = stokes.imshow(
                self.polparam_list[i].data, cmap=cmap, vmin=vmin, vmax=vmax
            )
            stokes.set_title(
                self.polparam_list[i].title
                + f" (avrg {avg:3.2f}+-{stdev:3.2f})",
                color="black",
            )
            stokes.set_xlabel("x [px]")
            stokes.set_ylabel("y [px]")
            stokes_fig.colorbar(
                mappable_stokes,
                ax=stokes,
                label=self.polparam_list[i].measure_unit,
                fraction=data_ratio * 0.05,
            )
        return image_fig, imageax, stokes_fig, stokesax

    def show_single_pol_images(self, cmap="Greys_r"):
        """Plots the four polarizations images.

        Args:
            cmap (str, optional): colormap for the plot. Defaults to "Greys_r".

        Returns:
            tuple: a (figure, axis) couple same as matplotlib.pyplot.subplots
        """
        data_ratio = self.data.shape[0] / self.data.shape[1]
        fig, ax = plt.subplots(2, 2, figsize=(9, 9), constrained_layout=True)
        ax = ax.ravel()
        polslist = [self.pol0, self.pol45, self.pol90, self.pol_45]
        for i, singlepolax in enumerate(ax):
            mappable = singlepolax.imshow(polslist[i].data, cmap=cmap)
            singlepolax.set_title(polslist[i].title)
            singlepolax.set_xlabel("x [px]")
            singlepolax.set_ylabel("y [px]")
            fig.colorbar(
                mappable,
                ax=singlepolax,
                label=polslist[i].measure_unit,
                fraction=data_ratio * 0.05,
                pad=0.01,
            )
        return fig, ax

    def show_demo_images(self, cmap="Greys_r"):
        """Plots the four demosaiced images.

        Args:
            cmap (str, optional): colormap for the plot. Defaults to "Greys_r".

        Returns:
            tuple: a (figure, axis) couple same as matplotlib.pyplot.subplots
        """
        if not self._is_demosaiced:
            error("Image is not yet demosaiced.")
        data_ratio = self.data.shape[0] / self.data.shape[1]
        fig, ax = plt.subplots(2, 2, figsize=(9, 9), constrained_layout=True)
        ax = ax.ravel()
        demo_images_list = self.demosaiced_images
        for i, single_demo_ax in enumerate(ax):
            mappable = single_demo_ax.imshow(
                demo_images_list[i],
                cmap=cmap,
                vmin=mean_minus_std(demo_images_list[i]),
                vmax=mean_plus_std(demo_images_list[i]),
            )
            single_demo_ax.set_title(
                f"Demosaiced image {list(self.angle_dic.keys())[list(self.angle_dic.values()).index(i)]}"
            )
            single_demo_ax.set_xlabel("x [px]")
            single_demo_ax.set_ylabel("y [px]")
            fig.colorbar(
                mappable,
                ax=single_demo_ax,
                label="DN",
                fraction=data_ratio * 0.05,
                pad=0.01,
            )
        return fig, ax

    def show_pol_param(
        self, polparam: PolParam, cmap="Greys_r", vmin=None, vmax=None
    ):
        """Plots a single polarization parameter given as input

        Args:
            polparam (PolParam): image PolParam containing the parameter to plot. Can be one among [self.I, self.Q, self.U, self.pB, self.AoLP, self.DoLP]
            cmap (str, optional): colormap for the plot. Defaults to "Greys_r".

        Returns:
            tuple: a (figure, axis) couple same as matplotlib.pyplot.subplots
        """
        data_ratio = self.data.shape[0] / self.data.shape[1]
        fig, ax = plt.subplots(dpi=200)
        if vmin is None:
            vmin = mean_minus_std(polparam.data)
        if vmax is None:
            vmax = mean_plus_std(polparam.data)
        mappable = ax.imshow(
            polparam.data,
            cmap=cmap,
            vmin=vmin,
            vmax=vmax,
        )
        ax.set_title(polparam.title)
        ax.set_xlabel("x [px]")
        ax.set_ylabel("y [px]")
        fig.colorbar(
            mappable,
            ax=ax,
            label=polparam.measure_unit,
            fraction=data_ratio * 0.05,
        )
        return fig, ax

    # ----------------------------------------------------------------
    # -------------------------- SAVING ------------------------------
    # ----------------------------------------------------------------

    def save_single_pol_images(
        self, filename: str, fixto: list[float, float] = None
    ) -> None:
        """Saves the four polarized images as fits files

        Args:
            filename (str): filename of the output image. The four images will be saved as filename_POLXX.fits
            fixto (list[float, float], optional): set the minimum and maximum value for the output images. Defaults to None.

        Raises:
            ValueError: an invalid file name is provided
        """
        polslist = [self.pol0, self.pol45, self.pol90, self.pol_45]
        filepath = Path(make_abs_and_create_dir(filename))
        if filepath.suffix != ".fits":
            raise ValueError("filename must be a valid file name, not folder.")
        group_filepath = filepath.joinpath(filepath.parent, filepath.stem)
        for single_pol in polslist:
            hdr = self.header.copy()
            hdr["POL"] = (single_pol.ID, "Micropolarizer orientation")
            if fixto:
                data = fix_data(single_pol.data, *fixto)
            else:
                data = single_pol.data
            hdu = fits.PrimaryHDU(
                data=data,
                header=hdr,
                do_not_scale_image_data=True,
                uint=False,
            )
            filename_with_ID = str(
                group_filepath.joinpath(
                    str(group_filepath) + "POL" + str(single_pol.ID) + ".fits"
                )
            )
            hdu.writeto(filename_with_ID, overwrite=True)
        info(f'All params successfully saved to "{filename}"')

    def save_param_as_fits(
        self,
        filename: str,
        polparam: PolParam,
        fixto: list[float, float] = None,
    ) -> None:
        """Saves chosen polarization parameter as a fits file

        Args:
            filename (str): filename of the output image.
            polparam (PolParam): polarization parameter to save. Can be one among [self.I, self.Q, self.U, self.pB, self.AoLP, self.DoLP]
            fixto (list[float, float], optional): set the minimum and maximum value for the output images. Defaults to None.

        Raises:
            ValueError: filename is not a valid .fits file
        """
        filepath = Path(make_abs_and_create_dir(filename))
        if filepath.suffix != ".fits":
            raise ValueError("filename must be a valid file name, not folder.")
        hdr = self.header.copy()
        hdr["PARAM"] = (str(polparam.title), "Polarization parameter")
        hdr["UNITS"] = (str(polparam.measure_unit), "Measure units")
        if fixto:
            data = fix_data(polparam.data, *fixto)
        else:
            data = polparam.data
        hdu = fits.PrimaryHDU(
            data=data,
            header=hdr,
            do_not_scale_image_data=True,
            uint=False,
        )
        filename_with_ID = str(
            filepath.joinpath(
                filepath.parent, filepath.stem + "_" + polparam.ID + ".fits"
            )
        )

        # filename = make_abs_and_create_dir(filename)
        # filename_with_ID = (
        #    filename.split(".")[-2] + "_" + polparam.ID + ".fits"
        # )
        hdu.writeto(filename_with_ID, overwrite=True)
        info(f'"{filename_with_ID}" {polparam.ID} successfully saved')

    def save_all_pol_params_as_fits(self, filename: str) -> None:
        """Saves the image and all polarization parameters as fits file with the same name

        Args:
            filename (str): filename of the output image. Will be saved as filename_[I, Q, U, pB, AoLP, DoLP].fits

        Raises:
            ValueError: filename is not a valid .fits file
        """
        filepath = Path(filename)
        if filepath.suffix != ".fits":
            raise ValueError("filename must be a valid file name, not folder.")
        filepath = Path(make_abs_and_create_dir(filename))
        group_filename = str(filepath.joinpath(filepath.parent, filepath.stem))
        for param in self.polparam_list:
            hdr = self.header.copy()
            hdr["PARAM"] = (str(param.title), "Polarization parameter")
            hdr["UNITS"] = (str(param.measure_unit), "Measure units")
            if param.fix_data:
                data = fix_data(param.data)
            else:
                data = param.data
            hdu = fits.PrimaryHDU(
                data=data,
                header=hdr,
                do_not_scale_image_data=True,
                uint=False,
            )
            filename_with_ID = group_filename + "_" + param.ID + ".fits"
            hdu.writeto(filename_with_ID, overwrite=True)
        info(f'All params successfully saved to "{group_filename}"')

    def save_demosaiced_images_as_fits(
        self, filename: str, fixto: list[float, float] = None
    ) -> None:
        """Saves the four demosaiced images as fits files

        Args:
            filename (str): filename of the output image. The four images will be saved as filename_POLXX.fits
            fixto (list[float, float], optional): set the minimum and maximum value for the output images. Defaults to None.

        Raises:
            ValueError: an invalid file name is provided
        """
        if not self._is_demosaiced:
            raise ValueError("Demosaiced images not yet calculated.")
        imageHdr = self.header.copy()
        filepath = Path(filename)
        if not filepath.suffix:
            raise ValueError("filename must be a valid file name, not folder.")
        filepath = Path(make_abs_and_create_dir(filename))
        group_filename = str(filepath.joinpath(filepath.parent, filepath.stem))
        for i, demo_image in enumerate(self.demosaiced_images):
            POL_ID = list(self.angle_dic.keys())[
                list(self.angle_dic.values()).index(i)
            ]
            imageHdr["POL"] = (int(POL_ID), "Micropolarizer orientation")
            if fixto:
                data = fix_data(demo_image, *fixto)
            else:
                data = demo_image
            hdu = fits.PrimaryHDU(
                data=data,
                header=imageHdr,
                do_not_scale_image_data=True,
                uint=False,
            )
            new_filename = group_filename + "_POL" + str(POL_ID) + ".fits"
            hdu.writeto(new_filename, overwrite=True)
        info(
            f'Demosaiced images successfully saved to "{group_filename}_POLX.fits"'
        )

    # ----------------------------------------------------------------
    # -------------------- DATA MANIPULATION -------------------------
    # ----------------------------------------------------------------
    def demosaic(self) -> MicropolImage:
        """Returns a demosaiced copy of the image with updated polarization parameters.

        Returns:
            MicropolImage: demosaiced image
        """
        self.demosaiced_images = demosaic(self.data, option=self.demosaic_mode)
        self.Stokes_vec = self._get_theo_Stokes_vec_components(
            self.demosaiced_images
        )
        self._update_stokes_derived_internal_dataclasses()
        self._is_demosaiced = True

        return self

    def rebin(self, binning: int) -> MicropolImage:
        """Rebins the micropolarizer array image, binned each
        binningxbinning. Sum bins by default.

        Args:
            binning (int): binning to perform. A value of n will be translated in a nxn binning.

        Raises:
            ValueError: negative binning provided

        Returns:
            MicropolImage: copy of the input image, rebinned.
        """
        if binning <= 0:
            raise ValueError(f"Negative binning {binning}x{binning}")
        rebinned_image = MicropolImage(self)
        rebinned_data = micropolarray_rebin(
            np.array(rebinned_image.data, dtype=float),
            *rebinned_image.data.shape,
            binning,
        )

        rebinned_image._set_data_and_Stokes(rebinned_data)

        return rebinned_image

    def congrid(self, newdim_y: int, newdim_x: int) -> MicropolImage:
        # Trim to nearest superpixel
        if (newdim_y % 2) or (newdim_x % 2):
            while newdim_y % 2:
                newdim_y = newdim_y - 1
            while newdim_x % 2:
                newdim_x = newdim_x - 1
            warning(
                f"New dimension was incompatible with superpixels. Trimmed to ({newdim_y}, {newdim_x})"
            )

        new_subdims = [int(newdim_y / 2), int(newdim_x / 2)]
        congridded_pol_images = [0.0] * 4
        for subimage_i, pol_subimage in enumerate(self.single_pol_subimages):
            congridded_pol_images[subimage_i] = congrid(
                pol_subimage, new_subdims
            )

        newdata = np.zeros(shape=(newdim_y, newdim_x))
        for pol_super_y in range(new_subdims[0]):
            for pol_super_x in range(new_subdims[1]):
                newdata[
                    pol_super_y * 2 : pol_super_y * 2 + 2,
                    pol_super_x * 2 : pol_super_x * 2 + 2,
                ] = [
                    [
                        congridded_pol_images[0][pol_super_y, pol_super_x],
                        congridded_pol_images[1][pol_super_y, pol_super_x],
                    ],
                    [
                        congridded_pol_images[2][pol_super_y, pol_super_x],
                        congridded_pol_images[3][pol_super_y, pol_super_x],
                    ],
                ]
        newimage = MicropolImage(newdata)
        return newimage

    def rotate(self, angle: float) -> MicropolImage:
        """Rotates an image of angle degrees, counter-clockwise."""

        single_pols = split_polarizations(self.data)
        for i in range(4):
            image = PILImage.fromarray(single_pols[i])
            image = image.rotate(angle)
            single_pols[i] = np.asarray(image)
        data = merge_polarizations(single_pols)

        return MicropolImage(data)

    def mask_occulter(
        self,
        y: int = PolarCam().occulter_pos_last[0],
        x: int = PolarCam().occulter_pos_last[1],
        r: int = PolarCam().occulter_pos_last[2],
        overoccult: int = 0,
    ) -> None:
        """Masks occulter for all image parameters

        Args:
            y (int, optional): Occulter y position. Defaults to PolarCam().occulter_pos_last[0].
            x (int, optional): Occulter x position. Defaults to PolarCam().occulter_pos_last[1].
            r (int, optional): Occulter radius. Defaults to PolarCam().occulter_pos_last[2].
            overoccult (int, optional): Pixels to overoccult. Defaults to 0.
            camera (_type_, optional): Camera image type. Defaults to PolarCam().

        Returns:
            None
        """
        # y, x, r = camera.occulter_pos_last

        r = r + overoccult

        self.data = roi_from_polar(
            self.data,
            (y, x),
            [r, 2 * np.max((self.height, self.width))],
        )
        self.single_pol_subimages = [
            roi_from_polar(
                data,
                (int(y / 2), int(x / 2)),
                [
                    int(r / 2),
                    2 * np.max((int(self.height / 2), int(self.width / 2))),
                ],
            )
            for data in self.single_pol_subimages
        ]
        self._update_single_pol_subimages()  # Updates polparam-like
        if self._is_demosaiced:
            self.demosaiced_images = [
                roi_from_polar(
                    data,
                    (y, x),
                    (r, 2 * np.max([self.height, self.width])),
                )
                for data in self.demosaiced_images
            ]
        for param in self.polparam_list:
            ratio = param.data.shape[0] / self.data.shape[0]
            param.data = roi_from_polar(
                param.data,
                (
                    int(y * ratio),
                    int(x * ratio),
                ),
                [
                    int(r * ratio),
                    2 * np.max((param.data.shape[0], param.data.shape[0])),
                ],
            )

    def shift(self, y: int, x: int) -> MicropolImage:
        """Shifts image by y, x pixels and fills with 0 the remaining space. Positive numbers for up/right shift and negative for down/left shift. Image is split into polarizations, each one is shifted, then they are merged again.

        Args:
            y (int): vertical shift in pix
            x (int): horizontal shift in pix

        Returns:
            MicropolImage: shifted image copied from the original
        """
        # newdata = shift(self.data, y, x)
        newdata = shift_micropol(self.data, y, x)
        newimage = MicropolImage(self)
        newimage._set_data_and_Stokes(newdata)

        return newimage

    def clean_hot_pixels(self, flagged_hot_pix_map: MicropolImage):
        """Returns a copy of the image with gaussian smeared pixels where flagged_hot_pix_map == 1.

        Args:
            flagged_hot_pix_map (MicropolImage): hot pixels map.

        Returns:
            MicropolImage: copy of the original image, gaussian smeared where flagged_hot_pix_map == 1
        """
        subimages = self.single_pol_subimages
        blurred_subimages = np.array(
            [
                scipy.ndimage.median_filter(subimage, size=2)
                for subimage in subimages
            ]
        )
        flagged_subimages = flagged_hot_pix_map.single_pol_subimages
        subimages = np.where(
            flagged_subimages == 1, blurred_subimages, subimages
        )

        newimage = MicropolImage(self)
        newimage._set_data_and_Stokes(merge_polarizations(subimages))
        return newimage

    # ----------------------------------------------------------------
    # ------------------------ OVERLOADING ---------------------------
    # ----------------------------------------------------------------
    def __add__(self, second) -> MicropolImage:
        if type(self) is type(second):
            newdata = self.data + second.data
            newimage = MicropolImage(self)
            newimage._set_data_and_Stokes(newdata)
            return newimage
        else:
            newdata = self.data + second
            return MicropolImage(newdata, angle_dic=self.angle_dic)

    def __sub__(self, second) -> MicropolImage:
        if type(self) is type(second):
            newdata = self.data - second.data
            newimage = MicropolImage(self)
            newimage._set_data_and_Stokes(newdata)
            return newimage
        else:
            newdata = self.data - second
            return MicropolImage(newdata, angle_dic=self.angle_dic)

    def __mul__(self, second) -> MicropolImage:
        if type(self) is type(second):
            newdata = self.data * second.data
            newimage = MicropolImage(self)
            newimage._set_data_and_Stokes(newdata)
            return newimage
        else:
            newdata = self.data * second
            return MicropolImage(newdata, angle_dic=self.angle_dic)

    def __truediv__(self, second) -> MicropolImage:
        if type(self) is type(second):
            newdata = np.divide(
                self.data, second.data, where=second.data != 0.0
            )
            newimage = MicropolImage(self)
            newimage._set_data_and_Stokes(newdata)
            return newimage
        else:
            # newdata = np.where(second != 0, self.data / second, 4096)
            newdata = np.divide(self.data, second, where=second != 0.0)
            return MicropolImage(newdata, angle_dic=self.angle_dic)


# provide shorter aliases
# PolarcamImage = MicropolImage
# MicropolImage = MicropolImage



import numpy as np
from astropy.io import fits
from micropolarray.cameras import PolarCam
from micropolarray.processing.nrgf import roi_from_polar
import sys
from numba import njit
from tqdm import tqdm
from pathlib import Path
from micropolarray.utils import make_abs_and_create_dir
from logging import warning, info


def calculate_chen_wan_lian_calibration(
    polarizer_orientations,
    filenames_list,
    micropol_phases_previsions,
    output_dir,
    occulter=True,
    dark_filename=None,
    flat_filename=None,
):
    """Performs calibration from Chen-Wang-Liang paper 2014

    Args:
        polarizer_orientations (list[float]): List of polarizer orienataions
        filenames_list (list[str]): List of filenames coupled with
        micropol_phases_previsions (list[float]): Previsions of the micropolarizer orientations inside a superpixel
        output_dir (str): output path for the calibration matrices
        occulter (bool, optional): wether to exclude the occulter area. Defaults to True.
        dark_filename (str, optional): path to the dark to be subtracted from the images. Defaults to None.
        flat_filename (str, optional): path to the dark to be subtracted from the images. Defaults to None.

    Raises:
        ValueError: polarizer orientation list and filenames list do not have the same lenght
        ValueError: any of 0,45,90,-45 polarizarions is not included in the polarizer orientation list
    """

    DEBUG = False

    if type(filenames_list) is not list:
        filenames_list = [filenames_list]

    pols_n = len(polarizer_orientations)
    filenames_n = len(filenames_list)
    if pols_n != filenames_n:
        raise ValueError("Polarizer orientations do not match filenames.")

    if not np.all(np.isin([0, 45, 90, -45], polarizer_orientations)):
        raise ValueError(
            "All (0, 45, 90, -45) pols must be included in the polarizer orientation array"
        )
    # Have to be sorted
    polarizer_orientations, filenames_list = (
        list(t)
        for t in zip(*sorted(zip(polarizer_orientations, filenames_list)))
    )

    angle_dic = {micropol_phases_previsions[i]: i for i in range(4)}

    with fits.open(filenames_list[0]) as file:
        data = file[0].data  # get data dimension
    occulter_flag = np.ones_like(data)  # 0 if not a occulted px, 1 otherwise
    if occulter:
        # Mean values from coronal observations 2022_12_03
        # (campagna_2022/mean_occulter_pos.py)

        occulter_y, occulter_x, occulter_r = PolarCam().occulter_pos_nov2022
        occulter_r += 10  # Better to overoccult

        occulter_flag = roi_from_polar(
            occulter_flag, [occulter_y, occulter_x], [0, occulter_r]
        )
        for super_y in range(0, occulter_flag.shape[0], 2):
            for super_x in range(0, occulter_flag.shape[1], 2):
                if np.any(
                    occulter_flag[super_y : super_y + 2, super_x : super_x + 2]
                ):
                    occulter_flag[
                        super_y : super_y + 2, super_x : super_x + 2
                    ] = 1
                    continue
    else:
        occulter_flag *= 0

    # Collect dark
    if dark_filename:
        with fits.open(dark_filename) as file:
            dark = file[0].data
    # Collect flat field, normalize it
    if flat_filename:
        with fits.open(flat_filename) as file:
            flat = file[0].data
    if flat_filename and dark_filename:
        flat -= dark  # correct flat too
        flat = np.where(flat > 0, flat, 1.0)
        if occulter:
            flat = np.where(occulter_flag, 1.0, flat)

    flat_max = np.max(flat, axis=(0, 1))
    normalized_flat = np.where(occulter_flag, 1.0, flat / flat_max)

    all_data_arr = [0.0] * filenames_n
    for idx, filename in enumerate(filenames_list):
        with fits.open(filename) as file:
            all_data_arr[idx] = file[0].data
            if dark_filename is not None:
                all_data_arr[idx] -= dark
                all_data_arr[idx] = np.where(
                    all_data_arr[idx] >= 0, all_data_arr[idx], 0.0
                )
            if flat_filename is not None:
                all_data_arr[idx] = np.where(
                    normalized_flat != 0,
                    all_data_arr[idx] / normalized_flat,
                    all_data_arr[idx],
                )

    all_data_arr = np.array(all_data_arr)
    _, height, width = all_data_arr.shape

    if DEBUG:
        warning("Running in DEBUG mode")
        height = int(height / 2)
        width = int(width / 2)
        all_data_arr = all_data_arr[:, 0:height, 0:width]

    # Stokes input vector: S = [pol] * [S_0, S_1, S_2, 1] * [width, height]
    S_input = np.zeros(shape=(filenames_n, 4, height, width))

    single_pol_images = np.array(
        [
            [
                all_data_arr[pol_idx, y::2, x::2]
                for y in range(2)
                for x in range(2)
            ]
            for pol_idx in range(pols_n)
        ]
    )

    sorted_single_pol_images = np.array(single_pol_images, copy=True)
    sorted_single_pol_images[:, 0, :, :] = single_pol_images[
        :, angle_dic[0], :, :
    ]
    sorted_single_pol_images[:, 1, :, :] = single_pol_images[
        :, angle_dic[45], :, :
    ]
    sorted_single_pol_images[:, 2, :, :] = single_pol_images[
        :, angle_dic[-45], :, :
    ]
    sorted_single_pol_images[:, 3, :, :] = single_pol_images[
        :, angle_dic[90], :, :
    ]

    build_S_input(sorted_single_pol_images, S_input)  # speeds up A LOT

    W_ideal = np.array(
        [
            [0.5, 0.5, 0.5, 0.5],
            [
                0.5 * np.cos(2.0 * np.deg2rad(micropol_phases_previsions[i]))
                for i in range(4)
            ],
            [
                0.5 * np.sin(2.0 * np.deg2rad(micropol_phases_previsions[i]))
                for i in range(4)
            ],
        ]
    ).T  # shape=(4,3)

    C_ij = np.zeros(shape=(pols_n, 4, 4, height, width))  # correction matrix
    d_ij = np.zeros(
        shape=(pols_n, height, width)
    )  # offsets, each element is equal in the superpixel so we can save one image with 4 pixel to save space

    if DEBUG:
        height = 8
        width = 8

    for pol_idx in tqdm(range(pols_n)):
        for super_y in tqdm(range(0, height, 2)):
            for super_x in range(0, width, 2):
                S = np.array([S_input[pol_idx, :, super_y, super_x]]).T

                S_inv_ij = np.linalg.pinv(S)
                P_ij = all_data_arr[
                    pol_idx, super_y : super_y + 2, super_x : super_x + 2
                ].reshape(4, 1)

                W_d = P_ij @ S_inv_ij
                # W_d = W_d.T
                W = W_d[:, :3]
                d = W_d[:, 3]
                try:
                    W_inv = np.linalg.pinv(W)
                except np.linalg.LinAlgError:
                    W_inv = 0.0 * W.T

                d_ij[
                    pol_idx, super_y : super_y + 2, super_x : super_x + 2
                ] = d.reshape(2, 2)

                C = W_ideal @ W_inv
                for i in range(2):
                    for j in range(2):
                        C_ij[pol_idx, :, :, super_y + i, super_x + j] = C

    C_ij = np.mean(C_ij, axis=0)
    d_ij = np.mean(d_ij, axis=0)

    output_dir = make_abs_and_create_dir(output_dir)
    for i in range(4):
        for j in range(4):
            hdu = fits.PrimaryHDU(data=C_ij[i, j])
            hdu.writeto(output_dir + f"/C{i}{j}.fits", overwrite=True)
    hdu = fits.PrimaryHDU(data=d_ij)
    hdu.writeto(output_dir + "/d_ij.fits", overwrite=True)


@njit
def build_S_input(single_pol_images, S_input):
    # ANGLES_LIST positions:
    #    0 <==> 0
    #    1 <==> 45
    #    2 <==> -45
    #    3 <==> 90
    pols_n = single_pol_images.shape[0]
    height = single_pol_images.shape[2] * 2
    width = single_pol_images.shape[3] * 2

    S_input[:, -1, :, :] += 1
    # Initialize stokes vector
    for pol_idx in range(pols_n):
        for super_y in range(0, height, 2):
            y = int(super_y / 2)
            for super_x in range(0, width, 2):
                x = int(super_x / 2)
                S_input[
                    pol_idx, 0, super_y : super_y + 2, super_x : super_x + 2
                ] = (
                    single_pol_images[pol_idx, 0, y, x]
                    + single_pol_images[pol_idx, 3, y, x]
                )
                S_input[
                    pol_idx, 1, super_y : super_y + 2, super_x : super_x + 2
                ] = (
                    single_pol_images[pol_idx, 0, y, x]
                    - single_pol_images[pol_idx, 3, y, x]
                )
                S_input[
                    pol_idx, 2, super_y : super_y + 2, super_x : super_x + 2
                ] = (
                    single_pol_images[pol_idx, 1, y, x]
                    - single_pol_images[pol_idx, 2, y, x]
                )


def chen_wan_liang_calibration(data, calibration_matrices_dir: str):
    """Calibrates the images using Chen-Wang-Liang 2014 paper calibration

    Args:
        data (np.array): data to be calibrated
        calibration_matrices_dir (str): path to the calibration matrices

    Returns:
        np.array: calibrated data
    """
    info("Applying Chen-Wan-Liang correction...")
    height, width = data.shape
    C_ij = np.zeros(shape=(4, 4, height, width))
    d_ij = np.zeros_like(data)
    corrected_data = np.zeros_like(data)
    ifov_corrected_data = np.zeros_like(data)
    correction_path = Path(calibration_matrices_dir)
    if not correction_path.is_absolute():  # suppose it is in cwd
        correction_path = correction_path.joinpath(
            Path().cwd(), correction_path
        )

    for i in range(4):
        for j in range(4):
            with fits.open(
                calibration_matrices_dir + f"/C{i}{j}.fits"
            ) as file:
                C_ij[i, j] = file[0].data
    with fits.open(calibration_matrices_dir + f"/d_ij.fits") as file:
        d_ij = file[0].data

    P_minus_d = data - d_ij

    for super_y in range(0, height, 2):
        for super_x in range(0, width, 2):
            mult = C_ij[:, :, super_y, super_x] @ P_minus_d[
                super_y : super_y + 2, super_x : super_x + 2
            ].reshape(4, 1)

            corrected_data[
                super_y : super_y + 2, super_x : super_x + 2
            ] = mult.reshape(2, 2)

    ifov_corrected_data = ifov_jitcorrect(corrected_data, height, width)

    return ifov_corrected_data


@njit
def ifov_jitcorrect(data, height, width):
    corrected_data = np.zeros(shape=(height, width))
    # correct IFOV error, skip first rows/cols
    for super_y in range(2, height, 2):
        for super_x in range(2, width, 2):
            for i in range(2):
                for j in range(2):
                    corrected_data[super_y + j, super_x + i] = 0.25 * (
                        data[super_y + j - 2, super_x + i - 2]
                        + data[super_y + j - 2, super_x + i]
                        + data[super_y + j, super_x + i - 2]
                        + data[super_y + j, super_x + i]
                    )
    return corrected_data



from distutils.core import setup

setup(
    name="micropolarray",
    description="Micro-Polarizer array and PolarCam images processing libraries",
    url="https://github.com/Hevil33/micropolarray_master",
    author="Hervé Haudemand",
    author_email="herve.haudemand@inaf.it",
    packages=[
        "micropolarray",
        "micropolarray.processing",
    ],  # name of the uppermost package directory
    # package_dir={"micropolarray": "./micropolarray"},
    project_urls={
        "Source Code": "https://github.com/Hevil33/micropolarray_master",
        "Documentation": "https://micropolarray.readthedocs.io/en/latest/",
    },
)



import sys
from functools import lru_cache
from logging import info

import matplotlib.pyplot as plt
import numpy as np
from numpy.lib.stride_tricks import as_strided
from scipy.optimize import curve_fit


def roi_from_polar(
    data: np.array,
    center: list = None,
    rho: list = None,
    theta=[0, 360],
    fill: float = 0.0,
    return_boolean=False,
) -> np.array:
    """Returns the input array in a circular selection, otherwise an arbitrary number. If a pixel is not in the selection the ENTIRE superpixel is considered out of selection. If return_boolean is True then a boolean array is returned instead (useful for mean/stdev operations).

    Args:
        data (np.array): input data
        center (list, optional): pixel coordinates of the circle center. Defaults to None (image center).
        rho (list, optional): radius to exclude. Defaults to None (center to image border).
        theta (list, optional): polar selection angle, 0 is horizonta, anti-clockwise direction. Defaults to [0, 360].
        fill (float, optional): number to fill the outer selection. Defaults to 0.0.
        return_boolean (bool, optional): if set to true, function returns a boolean array of the roi. Defaults to False.

    Returns:
        np.array: array containing the input data inside the selection, and fill otherwise
    """
    height, width = data.shape
    theta_min, theta_max = theta
    rho_min, rho_max = rho
    if center is None:
        center = [int(height / 2), int(width / 2)]
    if rho is None:
        rho_max = np.min([height - center[0], width - center[1]])
        rho = [0.0, rho_max]

    # make a map that is HALF THE SIZE, do condition, then resize to make all the superpixel outside selection
    rho_coords, phi_coords = map_polar_coordinates(
        int(height / 2),
        int(width / 2),
        tuple([int(center[0] / 2), int(center[1] / 2)]),
    )  # cast it to a tuple (which is hashable)

    theta_condition = np.logical_and(
        phi_coords >= theta_min, phi_coords < theta_max
    )
    rho_condition = np.logical_and(
        rho_coords > rho_min / 2, rho_coords <= rho_max / 2
    )  # half the radius because half the map
    condition = np.logical_and(rho_condition, theta_condition)

    # resize condition to correct shape
    condition = condition.repeat(2, axis=0).repeat(2, axis=1)

    if return_boolean:
        return np.where(condition, True, False)

    return np.where(condition, data, fill)


def tile_double(a):
    height, width = a.shape
    hs, ws = a.strides
    tiles = as_strided(a, (height, 2, width, 2), (hs, 0, ws, 0))
    return tiles.reshape(2 * height, 2 * width)


@lru_cache
def map_polar_coordinates(height, width, center):
    y_center, x_center = center

    i_list, j_list = np.arange(width), np.arange(height)
    x_coords, y_coords = np.meshgrid(i_list, j_list)
    # Map polar coordinates, 0 = horizontal dx, anti-clockwise angles
    rho_coords = np.sqrt(
        (x_coords - x_center) ** 2 + (y_coords - y_center) ** 2
    )
    phi_coords = (
        (np.arctan2(y_coords - y_center, x_coords - x_center) * 180 / np.pi)
        + 360
    ) % 360

    return rho_coords, phi_coords


def nrgf(
    data: np.array,
    y_center: int = None,
    x_center: int = None,
    rho_min: int = None,
    step: int = 1,
    phi_to_mean=[0.0, 360],
    output_phi=[0.0, 360],
) -> np.array:
    """
    Performs nrgf filtering on the image, starting from center and radius. Mean is performed between phi_to_mean, 0 is horizontal right, anti-clockwise.


    Args:
        data (np.array): input array
        y_center (int, optional): pixel y coordinate of the nrgf center. Defaults to None (image y center).
        x_center (int, optional): pixel x coordinate of the nrgf center. Defaults to (image x center).
        rho_min (int, optional): minimun radius in pixels to perform nrgf to. Defaults to None (radius 0).
        step (int, optional): step to which apply the nrgf from center, in pixels. Defaults to 1 pixel.
        phi_to_mean (list[float, float], optional): polar angle to calculate the mean value from. Defaults to [0, 360].
        output_phi (list[float, float], optional): polar angle to include in output data. Defaults to [0, 360].

    Returns:
        np.array: nrgf-filtered input data
    """
    height, width = data.shape

    if (y_center is None) or (x_center is None) or (rho_min is None):
        info("Calculating occulter position...")
        y_center, x_center, rho_min = find_occulter_position(data)
    center = [int(y_center), int(x_center)]

    newdata = np.zeros(shape=data.shape, dtype=float)
    i_list, j_list = np.arange(width), np.arange(height)
    x_coords, y_coords = np.meshgrid(i_list, j_list)
    # Map polar coordinates
    rho_coords, phi_coords = map_polar_coordinates(
        height, width, tuple(center)
    )  # cast it to a tuple (which is hashable)

    mean_phi_condition = np.logical_and(
        phi_coords >= phi_to_mean[0], phi_coords < phi_to_mean[1]
    )  # Exclude angle from mean

    out_phi_condition = np.logical_and(
        phi_coords >= output_phi[0], phi_coords < output_phi[1]
    )  # Exclude angle from filter

    rho_max = int(np.max(rho_coords))
    rho_step = step

    print("Applying nrgf filter...")
    for r in range(rho_min, rho_max, rho_step):
        rho_condition = np.logical_and(
            rho_coords > r, rho_coords <= r + rho_step
        )
        condition = np.logical_and(rho_condition, out_phi_condition)
        # condition = np.logical_and(rho_condition, mean_phi_condition)
        mean_condition = np.logical_and(rho_condition, mean_phi_condition)
        mean_over_ROI = np.mean(data, where=mean_condition)
        std_over_ROI = np.std(data, where=mean_condition)
        if std_over_ROI > 0:
            newdata = np.where(
                condition,
                (data - mean_over_ROI) / std_over_ROI,
                newdata,
            )
        else:
            newdata = np.where(condition, 0, newdata)

    return newdata


def find_occulter_position(
    data: np.array, method: str = "sigmoid", threshold: float = 4.0
):
    """Finds the occulter position of an image.

    Args:
        data (np.array): input data
        method (str, optional): Method to find occulter edges. If "sigmoid" it will try to fit four sigmoids at the image edges centers, inferring the occulter edges from the parameters. If "algo" it will start from the image edge center and infer the occulter position when DN[i] > threshold*mean(DN[:i]) Defaults to "sigmoid".
        threshold (float, optional): Threshold for the algo method. Defaults to 4.0.

    Raises:
        UnboundLocalError: couldn't converge

    Returns:
        list: occulter y, occulter x, occulter radius
    """
    # works if occulter is not entirely inside a single quadrant, fits
    # a sigmoid to find occulter bounds
    half_y = int(data.shape[0] / 2)
    half_x = int(data.shape[1] / 2)
    values_x_0 = np.flip(data[half_y, :half_x])
    values_x_1 = data[half_y, half_x:]
    values_y_0 = np.flip(data[:half_y, half_x])
    values_y_1 = data[half_y:, half_x]

    occulter_bounds = [0.0] * 4

    show = False
    if show:
        fig, ax = plt.subplots(2, 2, constrained_layout=True)
        ax = ax.ravel()
        titles = [
            "Lower vertical",
            "Upper vertical",
            "Right horizontal",
            "Left horizontal",
        ]

    if method == "sigmoid":
        for idx, half_array in enumerate(
            [values_y_0, values_y_1, values_x_0, values_x_1]
        ):
            # Artificial plateau after maximum
            # max_half_array = np.max(half_array)
            # for i, element in enumerate(half_array):
            #    if element == max_half_array:
            #        max_i = i
            #        break
            # half_array[max_i:] = max_half_array

            # x = np.arange(0, half_y, 1)
            x = np.arange(0, int(len(half_array / 2)), 1)

            params = [
                # np.mean(half_array[int(len(half_array) / 2) :]),
                np.max(half_array),
                # np.mean(half_array[: int(len(half_array) / 2)]),
                np.min(half_array),
                0.5,
                len(half_array) / 2,
            ]
            params, pcov = curve_fit(sigmoid, x, half_array, params)
            occulter_bounds[idx] = params[3]

            if show:
                axis = ax[idx]
                axis.scatter(x, half_array, c="grey")
                axis.plot(x, sigmoid(x, *params), c="black")
                axis.set_title(titles[idx])
                axis.set_xlabel("Pixels from center")
                axis.set_ylabel("DN")

        try:
            occulter_bounds[0] = half_y - occulter_bounds[0]
            occulter_bounds[1] = half_y + occulter_bounds[1]
            occulter_bounds[2] = half_x - occulter_bounds[2]
            occulter_bounds[3] = half_x + occulter_bounds[3]
        except UnboundLocalError:
            raise ValueError("Edges not found, try to change the threshold")

    elif method == "algo":
        # OLD, ALGORITM INSTEAD OF FITTING - less stable
        min_points = 5
        threshold = threshold

        for i, val in enumerate(values_x_0):
            if i > min_points:
                mean = np.mean(values_x_0[:i])
                if val > (threshold * mean):
                    occulter_start_x = half_x - i
                    break
        for i, val in enumerate(values_x_1):
            if i > min_points:
                mean = np.mean(values_x_1[:i])
                if val > (threshold * mean):
                    occulter_end_x = half_x + i
                    break
        for i, val in enumerate(values_y_0):
            if i > min_points:
                mean = np.mean(values_y_0[:i])
                if val > (threshold * mean):
                    occulter_start_y = half_y - i
                    break
        for i, val in enumerate(values_y_1):
            if i > min_points:
                mean = np.mean(values_y_1[:i])
                if val > (threshold * mean):
                    occulter_end_y = half_y + i
                    break
        try:
            occulter_bounds[0] = occulter_start_y
            occulter_bounds[1] = occulter_end_y
            occulter_bounds[2] = occulter_start_x
            occulter_bounds[3] = occulter_end_x
        except UnboundLocalError:
            raise ValueError(
                "ERROR: occulter edges not found, try to change the threshold"
            )
    elif method == "hugh":
        pass  # TODO
    if show:
        plt.show()

    y_center = int(np.mean([occulter_bounds[:2]]))
    x_center = int(np.mean([occulter_bounds[2:]]))
    radius = int(
        np.mean(
            [
                np.abs(occulter_bounds[1] - occulter_bounds[0]) / 2,
                np.abs(occulter_bounds[3] - occulter_bounds[2]) / 2,
            ]
        )
    )

    return [y_center, x_center, radius]


def remove_outliers_simple(original, neighbours=2):
    """EXPERIMENTAL DO NOT USE, for improving fitting on occulter position"""
    data = original.copy()
    for i, element in enumerate(data[neighbours:-neighbours]):
        median = np.median(data[i - neighbours : i + neighbours])
        std = np.median(np.abs(data[i - neighbours : i + neighbours] - median))
        if (element < median - 3 * std) or (element > median + 3 * std):
            print()
            print(element)
            print(data[neighbours:-neighbours])
            data[i] = median
            print(data[neighbours:-neighbours])
    return data

    median = np.median(data)

    median_deviation = np.abs(data - np.median(data))
    condition = (data < (median + 3 * median_deviation)) | (
        data > (median - 3 * median_deviation)
    )
    data = np.where(condition, data, median)
    return data
    extreme = 2
    outliers = []
    for i, element in enumerate(data[extreme:-extreme]):
        mean = np.mean([data[i], data[i + 1]])
        std = np.std([data[i], data[i + 1]])
        if (element > (mean + 3 * std)) or (element < (mean - 3 * std)):
            outliers.append(i + 1)
    data = np.delete(data, outliers)
    return data


def reject_outliers(data, m=2.0):
    d = np.abs(data - np.median(data))
    mdev = np.median(d)
    s = d / mdev if mdev else np.zeros(len(d))
    return data[s < m]


def sigmoid(x, max, min, slope, intercept):
    sigma = slope * (x - intercept)
    return max * np.exp(sigma) / (1 + np.exp(sigma)) + min



from logging import info

import numpy as np
from numba import jit, njit


def print_trimming_info(height, width, new_height, new_width):
    # print(
    #    f"Data trimmed to fit rebinning: ({height}, {width}) -> ({new_height}, {new_width})"
    # )
    print(
        "Data trimmed to fit rebinning: (",
        height,
        ",",
        width,
        ") -> (",
        new_height,
        ",",
        new_width,
        ")",
    )


def micropolarray_rebin(data: np.ndarray, height: int, width: int, binning=2):
    """Wrapper for the faster rebinning donw with numba. First deletes last row/column until binning is possible, then calls binning on the result shape.

    Args:
        data (np.ndarray): data to rebin
        height (int): lenght of first axis
        width (int): lenght of second axis
        binning (int, optional): Binning to be performed. Defaults to 2.

    Returns:
        ndarray: binned data, trimmed if necessary
    """
    trimmed = False
    new_height, new_width = trim_to_match_2xbinning(height, width, binning)
    new_height = int(new_height / binning)
    new_width = int(new_width / binning)
    new_data = micropolarray_jitrebin(data, new_height, new_width, binning)

    return new_data


@njit
def micropolarray_jitrebin_old(data, height, width, binning=2):
    """Fast rebinning function for the micropolarray image."""
    # Skip last row/columns until they are divisible by binning,
    # allows any binning
    trimmed = False
    new_height, new_width = trim_to_match_2xbinning(height, width, binning)
    new_height = int(new_height / binning)
    new_width = int(new_width / binning)
    new_data = np.zeros(shape=(new_height, new_width), dtype=float)
    for new_y in range(new_height):
        for new_x in range(new_width):
            for y_scaler in range((new_y % 2), (new_y % 2) + 2 * binning, 2):
                for x_scaler in range(
                    (new_x % 2), (new_x % 2) + 2 * binning, 2
                ):
                    i = (new_y - new_y % 2) * binning + y_scaler
                    j = (new_x - new_x % 2) * binning + x_scaler
                    new_data[new_y, new_x] += data[i, j]

    return new_data


@njit
def micropolarray_jitrebin(data, new_height, new_width, binning=2):
    """Fast rebinning function for the micropolarray image. Needs to be wrapped to print info."""
    new_data = np.zeros(shape=(new_height, new_width), dtype=float)
    for new_y in range(new_height):
        for new_x in range(new_width):
            for y_scaler in range((new_y % 2), (new_y % 2) + 2 * binning, 2):
                for x_scaler in range(
                    (new_x % 2), (new_x % 2) + 2 * binning, 2
                ):
                    i = (new_y - new_y % 2) * binning + y_scaler
                    j = (new_x - new_x % 2) * binning + x_scaler
                    new_data[new_y, new_x] += data[i, j]

    return new_data


def standard_rebin(data, binning: int) -> np.array:
    """Rebins the data, binned each binningxbinning.

    Args:
        image (np.array): data to be binned
        binning (int): binning to be applied. A value of 2 will result in a 2x2 binning (1 pixel is a sum of 4 neighbour pixels)

    Raises:
        KeyError: cannot divide image height/width by the binning value

    Returns:
        np.array: binned data
    """
    if (data.shape[0] % binning) or (data.shape[1] % binning):
        raise KeyError(
            "Invalid binning, must be divisor of both image height and width."
        )
    rebinned_data = standard_jitrebin(data, *data.shape, binning)
    return rebinned_data


@njit
def standard_jitrebin(data, height, width, binning=2):
    new_height, new_width = trim_to_match_2xbinning(height, width, binning)
    new_height = int(new_height / binning)
    new_width = int(new_width / binning)
    new_data = np.zeros(shape=(new_height, new_width))
    for new_y in range(new_height):
        for new_x in range(new_width):
            for y_scaler in range(binning):
                for x_scaler in range(binning):
                    new_data[new_y, new_x] += data[
                        binning * new_y + y_scaler,
                        binning * new_x + x_scaler,
                    ]

    return new_data


def trim_to_match_2xbinning(height: int, width: int, binning: int):
    """Deletes the last image pixels until superpixel binning is compatible with new dimensions

    Args:
        height (int): image height
        width (int): image width
        binning (int): image binning

    Returns:
        int, int: image new height and width
    """
    trimmed = False
    new_width = width
    new_height = height
    while new_width % (2 * binning):
        new_width -= 2
        trimmed = True
    while new_height % (2 * binning):
        new_height -= 2
        trimmed = True
    if trimmed:
        print_trimming_info(height, width, new_height, new_width)

    return new_height, new_width


def trim_to_match_binning(height, width, binning, verbose=True):
    """Deletes the last image pixels until simple binning is compatible with new dimensions

    Args:
        height (int): image height
        width (int): image width
        binning (int): image binning
        verbose (bool, optional): warns user of trimming. Defaults to True.

    Returns:
        int, int: image new height and width
    """
    trimmed = False
    new_height = height
    new_width = width
    while new_width % (binning):
        new_width -= 1
        trimmed = True
    while new_height % (binning):
        new_height -= 1
        trimmed = True
    if trimmed and verbose:
        print_trimming_info(height, width, new_height, new_width)
    return new_height, new_width



import glob
import multiprocessing as mp
import os
import re
import sys
import time
from logging import error, info, warning
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
from astropy.io import fits
from scipy.optimize import curve_fit
from tqdm import tqdm

from micropolarray.cameras import PolarCam
from micropolarray.processing.chen_wan_liang_calibration import ifov_jitcorrect
from micropolarray.processing.nrgf import (
    find_occulter_position,
    roi_from_polar,
)
from micropolarray.processing.rebin import (
    micropolarray_rebin,
    standard_rebin,
    trim_to_match_binning,
)
from micropolarray.utils import mean_plus_std

# Shape of the demodulation matrix
N_PIXELS_IN_SUPERPIX = 4
N_MALUS_PARAMS = 3


class Demodulator:
    """Demodulation class needed for MicropolImage
    demodulation."""

    def __init__(self, demo_matrices_path: str):
        self.n_pixels_in_superpix = N_PIXELS_IN_SUPERPIX
        self.n_malus_params = N_MALUS_PARAMS
        self.demo_matrices_path = demo_matrices_path

        self.mij = self._get_demodulation_tensor()

    def _get_demodulation_tensor(self):
        """Reads files "MIJ.fits" from path folder and returns a (3,4,y,x)
        numpy array representing the demodulation tensor.

        Args:
            binning (bool, optional): _description_. Defaults to False.

        Raises:
            FileNotFoundError: couldn't find the matrices in the specified path

        Returns:
            ndarray: (3, 4, *data.shape) array containing the demodulation tensor
        """

        if not os.path.exists(self.demo_matrices_path):
            raise FileNotFoundError("self.demo_matrices_path not found.")
        filenames_list = glob.glob(
            self.demo_matrices_path + os.path.sep + "*.fits"
        )

        with fits.open(filenames_list[0]) as firsthul:
            sample_data = np.array(firsthul[0].data)
        Mij = np.zeros(
            shape=(
                self.n_malus_params,
                self.n_pixels_in_superpix,
                sample_data.shape[0],
                sample_data.shape[1],
            ),
            dtype=float,
        )

        matches = 0
        for filename in filenames_list:
            if (
                re.search("[Mm][0-9]{2}", filename.split(os.path.sep)[-1])
                is not None
            ):  # Exclude files not matching m/Mij
                matches += 1
                i, j = re.search(
                    "[Mm][0-9]{2}", filename.split(os.path.sep)[-1]
                ).group()[
                    -2:
                ]  # Searches for pattern M/m + ij as last string before .fits
                i = int(i)
                j = int(j)
                with fits.open(filename) as hul:
                    Mij[i, j] = hul[0].data
        if matches != 12:
            raise ValueError(
                "Some matrices were not found in the selected folder. Check correct folder name and files pattern Mij.fits or mij.fits"
            )

        return Mij

    def show(self, vmin=-1, vmax=1, cmap="Greys") -> tuple:
        """Shows the demodulation tensor

        Args:
            vmin (int, optional): Minimum shown value. Defaults to -1.
            vmax (int, optional): Maximum shown value. Defaults to 1.
            cmap (str, optional): Colormap of the plot. Defaults to "Greys".

        Returns:
            tuple: fig, ax tuple as returned by matplotlib.pyplot.subplots
        """
        fig, ax = plt.subplots(
            3,
            4,
            dpi=300,
            figsize=(4, 3),
            constrained_layout=True,
            sharex="col",
            sharey="row",
        )
        for i in range(3):
            for j in range(4):
                mappable = ax[i, j].imshow(
                    self.mij[i, j], cmap=cmap, vmin=vmin, vmax=vmax
                )
                ax[i, j].set_title(rf"M$_{i}$$_{j}$")
        for ax in fig.get_axes():
            ax.label_outer()

        fig.subplots_adjust(right=0.8)
        cbar_ax = fig.add_axes([0.85, 0.15, 0.05, 0.7])
        fig.colorbar(mappable, cax=cbar_ax)

        return fig, ax

    def rebin(self, binning):  # TODO
        """DO NOT USE THIS, calculate the tensor from the binned images"""
        if (int(self.mij.shape[2] / binning) % 2) or (
            int(self.mij.shape[3] / binning) % 2
        ):
            raise ValueError(
                f"incorrect binning, resulting matrix would be {int(self.mij.shape[2] / binning)}x{int(self.mij.shape[3] / binning)} (not even values)."
            )
        new_demodulator = Demodulator(self.demo_matrices_path)
        new_mij = np.zeros(
            shape=(
                new_demodulator.n_malus_params,
                new_demodulator.n_pixels_in_superpix,
                int(new_demodulator.mij.shape[2] / binning),
                int(new_demodulator.mij.shape[3] / binning),
            )
        )
        for j in range(new_demodulator.n_malus_params):
            for i in range(new_demodulator.n_pixels_in_superpix):
                new_mij[j, i] = standard_rebin(
                    new_demodulator.mij[j, i], binning
                ) / (binning * binning)
        new_demodulator.mij = new_mij

        return new_demodulator

    def rot90(self, k=1):
        # NB: rotation could switch micropol positions inside superpixel, but for demo matrix all px in superpix are equal
        for i in range(self.n_malus_params):
            for j in range(self.n_pixels_in_superpix):
                self.mij[i, j] = np.rot90(self.mij[i, j], k=k)

    def flip(self, axis):
        for i in range(self.n_malus_params):
            for j in range(self.n_pixels_in_superpix):
                self.mij[i, j] = np.flip(self.mij[i, j], axis=axis)


def calculate_demodulation_tensor(
    polarizer_orientations,
    filenames_list,
    micropol_phases_previsions,
    gain,  #  needed for errors
    output_dir,
    binning=1,
    occulter=None,
    procs_grid=[4, 4],
    dark_filename=None,
    flat_filename=None,
    normalizing_S=None,
    DEBUG=False,
):
    """Calculates the demodulation tensor images and saves them. Requires a set of images with different polarizations to fit a Malus curve model.

    Args:
        polarizer_orientations (list[float]): List containing the orientations of the incoming light for each image.
        filenames_list (list[str]): List of input images filenames to read. Must include [0, 45, 90, -45].
        micropol_phases_previsions (list[float]): Previsions for the micropolarizer orientations required to initialize fit.
        gain (float): Detector [e-/DN], required to compute errors.
        output_dir (str): output folder to save matrix to.
        binning (int, optional): Output matrices binning. Defaults to 1 (no binning). Be warned that binning matrices AFTER calculation is an incorrect operation.
        occulter (list, optional): occulter y, x center and radius to exclude from calculations. Defaults to None.
        procs_grid ([int, int], optional): number of processors per side [Y, X], parallelization will be done in a Y x X grid. Defaults to [4,4] (16 procs in a 4x4 grid).
        dark_filename (str, optional): Dark image filename to correct input images. Defaults to None.
        flat_filename (str, optional): Flat image filename to correct input images. Defaults to None.
        normalizing_S (float, optional): maximum signal used to normalize single pixel signal. Defaults to None.

    Raises:
        ValueError: Raised if any among [0, 45, 90, -45] is not included in the input polarizations.

    Notes:
        In the binning process the sum of values is considered, which is ok because data is normalized over the maximum S before being fitted.
    """

    correct_ifov = False

    output_path = Path(output_dir)
    if not output_path.is_dir():  # create the path if it doesnt exist
        output_path.mkdir(parents=True)

    # polarizations = array of polarizer orientations
    # filenames_list = list of filenames
    firstcall = True
    if (normalizing_S is None) and (
        not np.all(np.isin([0, 45, 90, -45], polarizer_orientations))
    ):
        raise ValueError(
            "Each one among (0, 45, 90, -45) polarizations must be included in the polarizer orientation array"
        )  # for calculating normalizing_S
    # Have to be sorted
    polarizer_orientations, filenames_list = (
        list(t)
        for t in zip(*sorted(zip(polarizer_orientations, filenames_list)))
    )
    micropol_phases_previsions = np.array(micropol_phases_previsions)
    rad_micropol_phases_previsions = np.deg2rad(micropol_phases_previsions)

    # Flag occulter position to not be fitted, expand to superpixel.
    with fits.open(filenames_list[0]) as file:
        data = file[0].data  # get data dimension

    # Count binning before dimensions
    data = np.array(data, dtype=float)
    data = micropolarray_rebin(data, *data.shape, binning=binning)
    height, width = data.shape

    occulter_flag = np.ones_like(data)  # 0 if not a occulted px, 1 otherwise
    if occulter:
        # info("Cleaning occulter...")
        ## Mean values from coronal observations 2022_12_03
        ## (campagna_2022/mean_occulter_pos.py)
        # occulter_y, occulter_x, occulter_r = PolarCam().occulter_pos_last
        # overoccult = 16
        # overoccult = 0
        # occulter_r = occulter_r + overoccult

        # Match binning if needed
        binned_occulter = [int(val / binning) for val in occulter]
        occulter_y, occulter_x, occulter_r = binned_occulter

        occulter_flag = roi_from_polar(
            occulter_flag, [occulter_y, occulter_x], [0, occulter_r]
        )
        for super_y in range(0, occulter_flag.shape[0], 2):
            for super_x in range(0, occulter_flag.shape[1], 2):
                if np.any(
                    occulter_flag[super_y : super_y + 2, super_x : super_x + 2]
                ):
                    occulter_flag[
                        super_y : super_y + 2, super_x : super_x + 2
                    ] = 1
                    continue
    else:
        occulter_flag *= 0

    # Collect dark
    if dark_filename:
        with fits.open(dark_filename) as file:
            dark = np.array(file[0].data, dtype=np.float)
            dark = micropolarray_rebin(dark, *dark.shape, binning)
    # Collect flat field, normalize it
    if flat_filename:
        with fits.open(flat_filename) as file:
            flat = np.array(file[0].data, dtype=np.float)
        if correct_ifov:
            flat = ifov_jitcorrect(flat, *flat.shape)
        flat = micropolarray_rebin(flat, *flat.shape, binning)
        # flat_max = np.max(flat, axis=(0, 1))
        flat_max = mean_plus_std(flat, stds_n=1)
    if flat_filename and dark_filename:
        flat -= dark  # correct flat too
        flat = np.where(flat > 0, flat, 1.0)
        if occulter:
            flat = np.where(occulter_flag, 1.0, flat)
        # flat_max = np.max(flat, axis=(0, 1))
        flat_max = mean_plus_std(flat, stds_n=1)
    if flat_filename:
        normalized_flat = np.where(occulter_flag, 1.0, flat / flat_max)

    # collect data
    all_data_arr = [0.0] * len(filenames_list)
    info("Collecting data from files...")
    for idx, filename in enumerate(filenames_list):
        with fits.open(filename) as file:
            all_data_arr[idx] = np.array(file[0].data, dtype=float)
            if correct_ifov:
                all_data_arr[idx] = ifov_jitcorrect(
                    all_data_arr[idx], *all_data_arr[idx].shape
                )
            all_data_arr[idx] = micropolarray_rebin(
                all_data_arr[idx], *all_data_arr[idx].shape, binning
            )
            if dark_filename is not None:
                all_data_arr[idx] -= dark
                all_data_arr[idx] = np.where(
                    all_data_arr[idx] >= 0, all_data_arr[idx], 0.0
                )
            if flat_filename is not None:
                all_data_arr[idx] = np.divide(
                    all_data_arr[idx],
                    normalized_flat,
                    where=normalized_flat != 0.0,
                )
    all_data_arr = np.array(all_data_arr)

    if DEBUG:
        procs_grid = [1, 1]

    # parallelize into a procs_per_size x procs_per_size grid
    info("Splitting into subdomains to parallelize...")
    chunks_n_y, chunks_n_x = procs_grid
    chunk_size_y = int(height / chunks_n_y)
    chunk_size_x = int(width / chunks_n_x)
    if (chunk_size_x % 2) or (chunk_size_y % 2):
        raise ValueError(
            f"cant decompose into a {procs_grid[0]}x{procs_grid[1]} grid (odd side grid {chunk_size_y}x{chunk_size_x}). Try changing the number of processors."
        )
    splitted_data = np.zeros(
        shape=(
            chunks_n_y * chunks_n_x,
            len(polarizer_orientations),
            chunk_size_y,
            chunk_size_x,
        )
    )
    splitted_occulter = np.zeros(
        shape=(chunks_n_y * chunks_n_x, chunk_size_y, chunk_size_x)
    )
    for i in range(chunks_n_y):
        for j in range(chunks_n_x):
            splitted_data[i + chunks_n_y * j] = np.array(
                all_data_arr[
                    :,
                    i * (chunk_size_y) : (i + 1) * chunk_size_y,
                    j * (chunk_size_x) : (j + 1) * chunk_size_x,
                ]
            )  # shape = (chunks_n*chunks_n, len(filenames_list), chunk_size_y, chunk_size_x)
            splitted_occulter[i + chunks_n_y * j] = np.array(
                occulter_flag[
                    i * (chunk_size_y) : (i + 1) * chunk_size_y,
                    j * (chunk_size_x) : (j + 1) * chunk_size_x,
                ]
            )  # shape = (chunks_n*chunks_n, chunk_size_y, chunk_size_x)

    if normalizing_S is None:
        info("Calculating normalization...")
        S_max = np.zeros(
            shape=(height, width)
        )  # tk_sum = tk_0 + tk_45 + tk_90 + tk_45
        for pol, image in zip(polarizer_orientations, all_data_arr):
            if pol in [0, 90, 45, -45]:
                S_max += 0.5 * image
        # Normalizing S, has a spike of which maximum is taken
        bins = 1000
        histo = np.histogram(S_max, bins=bins)
        maxvalue = np.max(histo[0])
        index = np.where(histo[0] == maxvalue)[0][0]
        normalizing_S = (
            histo[1][index] + histo[1][index + 1] + histo[1][index - 1]
        ) / 3
        # normalizing_S = np.max(S_max) # old

        # ----------------------------------------------
        # fit gaussian to S for normalization signal
        def gauss(x, norm, x_0, sigma):
            return norm * np.exp(-((x - x_0) ** 2) / (4 * sigma**2))

        hist_roi = 10  # bins around max value
        xvalues = np.array(histo[1])[index - hist_roi : index + hist_roi]
        yvalues = np.array(histo[0])[index - hist_roi : index + hist_roi]
        yvalues_sum = np.sum(yvalues)

        yvalues = yvalues / yvalues_sum
        xvalues = np.array(
            [value + (xvalues[1] - xvalues[0]) / 2 for value in xvalues]
        )  # shift each bin to center
        prediction = [
            yvalues[int(len(yvalues) / 2)],  # normalization
            xvalues[int(len(xvalues) / 2)],  # center
            xvalues[int(len(xvalues) / 2) + int(hist_roi / 2)]
            - xvalues[int(len(xvalues) / 2)],  # sigma
        ]
        params, cov = curve_fit(
            gauss,
            xvalues,
            yvalues,
            prediction,
        )
        normalizing_S = (
            params[1] + 4 * params[2]
        )  # center of gaussian + 2sigma
        # 3sigma -> P = 2.7e-3 outliers
        # 4sigma -> P = 6.3e-5 outliers
        # ----------------------------------------------
    else:
        normalizing_S *= binning * binning  # account binning

    # Debug
    if False:
        index = 0
        histo_0 = np.histogram(all_data_arr[index], bins=1000)
        fig, ax = plt.subplots(figsize=(9, 9))
        ax.stairs(histo_0[0], histo_0[1], label="sample image")
        ax.stairs(histo[0], histo[1], label=f"S, max = {np.max(S_max)}")
        ax.axvline(normalizing_S, color="red", label="normalizing_S")
        # ax.plot(
        #    xvalues,
        #    gauss(xvalues, params[0] * yvalues_sum, params[1], params[2]),
        #    label="Fitted curve for normalizing S",
        # )
        ax.set_title(f"Prepol at {polarizer_orientations[index]} deg")
        ax.set_xlabel("S [DN]")
        ax.set_ylabel("Counts")
        ax.legend()
        plt.show()
        sys.exit()

    args = (
        [
            splitted_data[i],
            normalizing_S,
            splitted_occulter[i],
            polarizer_orientations,
            rad_micropol_phases_previsions,
            gain,
            DEBUG,
        ]
        for i in range(chunks_n_y * chunks_n_x)
    )

    starting_time = time.perf_counter()
    loc_time = time.strftime("%H:%M:%S  (%Y/%m/%d)", time.localtime())
    info(
        f"Starting parallel calculation ({procs_grid[0]}x{procs_grid[1]}) processors"
    )

    if procs_grid != [1, 1]:
        try:
            with mp.Pool(processes=chunks_n_y * chunks_n_x) as p:
                result = p.starmap(
                    compute_demodulation_by_chunk,
                    args,
                )
        except:
            error("Fit not found")
            ending_time = time.perf_counter()

            info(f"Elapsed : {(ending_time - starting_time)/60:3.2f} mins")
            sys.exit()

    else:
        arglist = [arg for arg in args]
        result = [[0.0, 0.0]] * chunks_n_y * chunks_n_x

        for i in range(chunks_n_y * chunks_n_x):
            result[i] = compute_demodulation_by_chunk(*arglist[i])

    loc_time = time.strftime("%H:%M:%S (%Y/%m/%d)", time.localtime())
    info(f"Ending parallel calculation")

    ending_time = time.perf_counter()
    info(f"Elapsed : {(ending_time - starting_time)/60:3.2f} mins")

    result = np.array(result, dtype=object)
    m_ij = np.zeros(
        shape=(N_MALUS_PARAMS, N_PIXELS_IN_SUPERPIX, height, width)
    )
    tks = np.zeros(shape=(height, width))
    efficiences = np.zeros(shape=(height, width))
    phases = np.zeros(shape=(height, width))

    for i in range(chunks_n_y):
        for j in range(chunks_n_x):
            m_ij[
                :,
                :,
                i * (chunk_size_y) : (i + 1) * chunk_size_y,
                j * (chunk_size_x) : (j + 1) * chunk_size_x,
            ] = result[i + chunks_n_y * j, 0].reshape(
                N_MALUS_PARAMS,
                N_PIXELS_IN_SUPERPIX,
                chunk_size_y,
                chunk_size_x,
            )
            tks[
                i * (chunk_size_y) : (i + 1) * chunk_size_y,
                j * (chunk_size_x) : (j + 1) * chunk_size_x,
            ] = result[i + chunks_n_y * j, 1].reshape(
                chunk_size_y, chunk_size_x
            )
            efficiences[
                i * (chunk_size_y) : (i + 1) * chunk_size_y,
                j * (chunk_size_x) : (j + 1) * chunk_size_x,
            ] = result[i + chunks_n_y * j, 2].reshape(
                chunk_size_y, chunk_size_x
            )
            phases[
                i * (chunk_size_y) : (i + 1) * chunk_size_y,
                j * (chunk_size_x) : (j + 1) * chunk_size_x,
            ] = result[i + chunks_n_y * j, 3].reshape(
                chunk_size_y, chunk_size_x
            )

    phases = np.rad2deg(phases)

    if DEBUG:
        # prevents overwriting
        sys.exit()

    if not os.path.exists(output_dir):
        p = Path(output_dir)
        # p.mkdir(parent=True, exist_ok=True)
        os.makedirs(output_dir)

    output_str = str(output_path)
    for i in range(N_MALUS_PARAMS):
        for j in range(N_PIXELS_IN_SUPERPIX):
            hdu = fits.PrimaryHDU(data=m_ij[i, j])
            hdu.writeto(output_str + f"/M{i}{j}.fits", overwrite=True)
    hdu = fits.PrimaryHDU(data=tks)
    hdu.writeto(output_str + "/transmittancies.fits", overwrite=True)
    hdu = fits.PrimaryHDU(data=efficiences)
    hdu.writeto(output_str + "/efficiences.fits", overwrite=True)
    hdu = fits.PrimaryHDU(data=phases)
    hdu.writeto(output_str + "/phases.fits", overwrite=True)

    info("Demodulation matrices and fit data successfully saved!")

    return


def compute_demodulation_by_chunk(
    splitted_dara_arr,
    normalizing_S,
    splitted_occulter_flag,
    polarizer_orientations,
    rad_micropol_phases_previsions,
    gain,
    DEBUG,
):
    """Utility function to parallelize calculations."""
    N_MALUS_PARAMS = 3
    N_PIXELS_IN_SUPERPIX = 4
    # Preemptly compute the theoretical demo matrix to save time
    theo_modulation_matrix = np.array(
        [
            [0.5, 0.5, 0.5, 0.5],
            [
                0.5 * np.cos(2.0 * rad_micropol_phases_previsions[i])
                for i in range(N_PIXELS_IN_SUPERPIX)
            ],
            [
                0.5 * np.sin(2.0 * rad_micropol_phases_previsions[i])
                for i in range(N_PIXELS_IN_SUPERPIX)
            ],
        ],
        dtype=float,
    )
    theo_modulation_matrix = theo_modulation_matrix.T
    theo_demodulation_matrix = np.linalg.pinv(theo_modulation_matrix)

    num_of_points, height, width = splitted_dara_arr.shape
    rad_micropol_phases_previsions = np.array(
        rad_micropol_phases_previsions, dtype=float
    )
    polarizations_rad = np.deg2rad(polarizer_orientations)
    tk_prediction = 0.5
    efficiency_prediction = 0.4

    normalized_splitted_data = splitted_dara_arr / normalizing_S
    all_zeros = np.zeros(shape=(num_of_points))

    # Checked errors
    sigma_S2 = np.sqrt(0.5 * normalizing_S / gain)
    normalizing_S2 = normalizing_S * normalizing_S
    # pix_DN_sigma = np.sqrt(
    #    splitted_dara_arr / (gain * normalizing_S2)
    #    + sigma_S2
    #    * (splitted_dara_arr * splitted_dara_arr)
    #    / (normalizing_S2 * normalizing_S2)
    # )
    pix_DN_sigma = (
        np.sqrt(splitted_dara_arr / gain) / normalizing_S
    )  # poisson error on the photoelectrons
    # pix_DN_sigma = np.sqrt(normalized_splitted_data / gain)

    m_ij = np.zeros(
        shape=(N_MALUS_PARAMS, N_PIXELS_IN_SUPERPIX, height, width)
    )  # demodulation matrix
    tk_data = np.ones(shape=(height, width)) * tk_prediction
    eff_data = np.ones(shape=(height, width)) * efficiency_prediction
    phase_data = np.zeros(shape=(height, width))
    superpix_params = np.zeros(shape=(N_PIXELS_IN_SUPERPIX, N_MALUS_PARAMS))

    predictions = np.zeros(shape=(N_PIXELS_IN_SUPERPIX, N_MALUS_PARAMS))
    predictions[:, 0] = tk_prediction  # Throughput prediction
    predictions[:, 1] = efficiency_prediction  # Efficiency prediction
    predictions[:, 2] = rad_micropol_phases_previsions  # Angle prediction

    bounds = np.zeros(shape=(N_PIXELS_IN_SUPERPIX, 2, N_MALUS_PARAMS))
    bounds[:, 0, 0], bounds[:, 1, 0] = 0.1, 0.9999999  # Throughput bounds

    # bounds[:, 0, 0], bounds[:, 1, 0] = (
    #    0.1,
    #    2.0,
    # )  # tk is multiplied by 0.5 so it can actually be > 1 and still physical
    bounds[:, 0, 1], bounds[:, 1, 1] = 0.1, 0.9999999  # Efficiency bounds
    bounds[:, 0, 2] = rad_micropol_phases_previsions - 15  # Lower angle bounds
    bounds[:, 1, 2] = rad_micropol_phases_previsions + 15  # Upper angle bounds

    # Fit for each superpixel. Use theoretical demodulation matrix for
    # occulter if present.
    if DEBUG:
        x_start, x_end = 100, 110
        x_start, x_end = 0, 2
        y_start, y_end = 100, 110
        y_start, y_end = 0, 2
    else:
        y_start, y_end = 0, height
        x_start, x_end = 0, width
    milestones = [
        int(height / 4),
        int(height / 4) + 1,
        int(height / 2),
        int(height / 2) + 1,
        int(3 * height / 4),
        int(3 * height / 4) + 1,
        int(height),
        int(height) + 1,
    ]  # used for printing progress
    for super_y in range(y_start, y_end, 2):
        if super_y in milestones:
            print(f"Thread at {super_y / height*100:.2f} %", flush=True)
        for super_x in range(x_start, x_end, 2):
            if not (
                np.any(
                    splitted_occulter_flag[
                        super_y : super_y + 2, super_x : super_x + 2
                    ]
                )
            ):
                normalized_superpix_arr = normalized_splitted_data[
                    :, super_y : super_y + 2, super_x : super_x + 2
                ].reshape(num_of_points, N_PIXELS_IN_SUPERPIX)

                sigma_pix = pix_DN_sigma[
                    :, super_y : super_y + 2, super_x : super_x + 2
                ].reshape(num_of_points, N_PIXELS_IN_SUPERPIX)
                sigma_pix = np.where(sigma_pix != 0.0, sigma_pix, 1.0e-5)

                for pixel_num in range(N_PIXELS_IN_SUPERPIX):
                    if np.array_equal(
                        normalized_superpix_arr[:, pixel_num], all_zeros
                    ):  # catch bad pixels
                        fit_success = False
                        break
                    try:
                        (
                            superpix_params[pixel_num],
                            superpix_pcov,
                        ) = curve_fit(
                            Malus,
                            polarizations_rad,
                            normalized_superpix_arr[:, pixel_num],
                            predictions[pixel_num],
                            sigma=sigma_pix[:, pixel_num],
                            bounds=bounds[pixel_num],
                        )
                        fit_success = True
                    except RuntimeError:
                        fit_success = False
                        break

                if DEBUG:  # DEBUG
                    colors = ["blue", "orange", "green", "red"]
                    fig, ax = plt.subplots(
                        figsize=(9, 9), constrained_layout=True
                    )
                    for i in range(4):
                        ax.errorbar(
                            np.rad2deg(polarizations_rad),
                            normalized_superpix_arr[:, i],
                            yerr=sigma_pix[:, i],
                            xerr=[1.0] * len(polarizations_rad),
                            label=f"points {i}",
                            fmt="k-",
                            color=colors[i],
                        )
                        min = np.min(polarizations_rad)
                        # min = 225
                        max = np.max(polarizations_rad)
                        # max = 405
                        x = np.arange(min, max, (max - min) / 100)
                        ax.plot(
                            np.rad2deg(x),
                            Malus(x, *superpix_params[i]),
                            label=f"t = {superpix_params[i,0]:2.2f}, e = {superpix_params[i, 1]:2.2f}, phi = {np.rad2deg(superpix_params[i, 2]):2.2f}",
                            color=colors[i],
                        )
                        ax.set_title(
                            f"super_y = {super_y}, super_x = {super_x},"
                        )
                        ax.set_xlabel("Prepolarizer orientations [deg]")
                        ax.set_ylabel("signal / S")

                    plt.legend()
                    plt.show()

                if not fit_success:
                    for i in range(2):
                        for j in range(2):
                            m_ij[
                                :, :, super_y + i, super_x + j
                            ] = theo_demodulation_matrix
                    continue

                # Compute modulation matrix and its inverse
                t = superpix_params[:, 0]
                eff = superpix_params[:, 1]
                phi = superpix_params[:, 2]

                modulation_matrix = np.array(
                    [
                        0.5 * t,
                        0.5 * t * eff * np.cos(2.0 * phi),
                        0.5 * t * eff * np.sin(2.0 * phi),
                    ],
                    dtype=float,
                )
                modulation_matrix = modulation_matrix.T
                demodulation_matrix = np.linalg.pinv(modulation_matrix)

                if DEBUG:
                    print()
                    print("MOD")
                    print(modulation_matrix)
                    print(theo_modulation_matrix)

                    print()
                    print("DEMOD")
                    print(demodulation_matrix)
                    print(theo_demodulation_matrix)

                    print()
                    print("params")
                    print(t)
                    print(eff)
                    print(phi)
                    print("---")

                # Remove matrices with big numbers
                if np.any(demodulation_matrix > 100) or np.any(
                    demodulation_matrix < -100
                ):
                    demodulation_matrix = theo_demodulation_matrix

                for i in range(2):
                    for j in range(2):
                        m_ij[
                            :, :, super_y + i, super_x + j
                        ] = demodulation_matrix

                tk_data[
                    super_y : super_y + 2, super_x : super_x + 2
                ] = np.array(t, dtype=float).reshape(2, 2)
                eff_data[
                    super_y : super_y + 2, super_x : super_x + 2
                ] = np.array(eff, dtype=float).reshape(2, 2)
                phase_data[
                    super_y : super_y + 2, super_x : super_x + 2
                ] = np.array(phi, dtype=float).reshape(2, 2)

            else:  # pixel is in occulter region
                for i in range(2):
                    for j in range(2):
                        m_ij[
                            :, :, super_y + i, super_x + j
                        ] = theo_demodulation_matrix
                phase_data[
                    super_y : super_y + 2, super_x : super_x + 2
                ] = rad_micropol_phases_previsions.reshape(2, 2)
                tk_data[
                    super_y : super_y + 2, super_x : super_x + 2
                ] = np.array(
                    [
                        [tk_prediction, tk_prediction],
                        [tk_prediction, tk_prediction],
                    ],
                    dtype=float,
                )
                eff_data[
                    super_y : super_y + 2, super_x : super_x + 2
                ] = np.array(
                    [
                        [efficiency_prediction, efficiency_prediction],
                        [efficiency_prediction, efficiency_prediction],
                    ],
                    dtype=float,
                )

    m_ij_chunk = m_ij

    return m_ij_chunk, tk_data, eff_data, phase_data


def Malus(angle, throughput, efficiency, phase):
    # original
    modulated_efficiency = efficiency * (
        np.cos(2.0 * phase) * np.cos(2.0 * angle)
        + np.sin(2.0 * phase) * np.sin(2.0 * angle)
    )
    return 0.5 * throughput * (1.0 + modulated_efficiency)
